# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import json
import logging
import time
import typing

from decimal import Decimal
from solana.publickey import PublicKey
from solana.rpc.types import MemcmpOpts, DataSliceOpts, RPCResponse

from .constants import SOL_DECIMAL_DIVISOR
from .context import Context
from .encoding import decode_binary, encode_binary


# # 🥭 AccountInfo class
#
class AccountInfo:
    def __init__(
        self,
        address: PublicKey,
        executable: bool,
        lamports: Decimal,
        owner: PublicKey,
        rent_epoch: Decimal,
        data: bytes,
    ) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.address: PublicKey = address
        self.executable: bool = executable
        self.lamports: Decimal = lamports
        self.owner: PublicKey = owner
        self.rent_epoch: Decimal = rent_epoch
        self.data: bytes = data

    @property
    def sols(self) -> Decimal:
        return self.lamports / SOL_DECIMAL_DIVISOR

    def encoded_data(self) -> typing.Sequence[str]:
        return encode_binary(self.data)

    def save_json(self, filename: str) -> None:
        data = {
            "address": str(self.address),
            "executable": self.executable,
            "lamports": str(self.lamports),
            "owner": str(self.owner),
            "rent_epoch": str(self.rent_epoch),
            "data": encode_binary(self.data),
        }
        with open(filename, "w") as json_file:
            json.dump(data, json_file, indent=4)

    def __str__(self) -> str:
        return f"""« AccountInfo [{self.address}]:
    Owner: {self.owner}
    Executable: {self.executable}
    Lamports: {self.lamports} / Sols: {self.sols}
    Rent Epoch: {self.rent_epoch}
    Data Length: {len(self.data)}
»"""

    def __repr__(self) -> str:
        return f"{self}"

    @staticmethod
    def load(context: Context, address: PublicKey) -> typing.Optional["AccountInfo"]:
        result = context.client.get_account_info(address)
        if result["value"] is None:
            return None

        return AccountInfo._from_response_values(result["value"], address)

    @staticmethod
    def load_json(filename: str) -> "AccountInfo":
        with open(filename) as json_file:
            accountinfo_data = json.load(json_file)
            address: PublicKey = PublicKey(accountinfo_data["address"])
            executable: bool = accountinfo_data["executable"]
            lamports: Decimal = Decimal(accountinfo_data["lamports"])
            owner: PublicKey = PublicKey(accountinfo_data["owner"])
            rent_epoch: Decimal = Decimal(accountinfo_data["lamports"])
            data: bytes = decode_binary(accountinfo_data["data"])
            return AccountInfo(address, executable, lamports, owner, rent_epoch, data)

    @staticmethod
    def load_multiple(
        context: Context, addresses: typing.Sequence[PublicKey]
    ) -> typing.Sequence["AccountInfo"]:
        # This is a tricky one to get right.
        # Some errors this can generate:
        #  413 Client Error: Payload Too Large for url
        #  Error response from server: 'Too many inputs provided; max 100', code: -32602
        chunk_size: int = int(context.gma_chunk_size)
        sleep_between_calls: float = float(context.gma_chunk_pause)
        multiple: typing.List[AccountInfo] = []
        chunks: typing.Sequence[
            typing.Sequence[PublicKey]
        ] = AccountInfo._split_list_into_chunks(addresses, chunk_size)
        for counter, chunk in enumerate(chunks):
            result: typing.Sequence[
                typing.Dict[str, typing.Any]
            ] = context.client.get_multiple_accounts([*chunk])
            response_value_list = zip(result, chunk)
            for index, pair in enumerate(response_value_list):
                if pair[0] is None:
                    raise Exception(
                        f"Failed to fetch account {chunk[index]} at index {index}"
                    )
                multiple += [AccountInfo._from_response_values(pair[0], pair[1])]
            if (sleep_between_calls > 0.0) and (counter < (len(chunks) - 1)):
                time.sleep(sleep_between_calls)

        return multiple

    @staticmethod
    def load_by_program(
        context: Context,
        pubkey: typing.Union[str, PublicKey],
        data_slice: typing.Optional[DataSliceOpts] = None,
        data_size: typing.Optional[int] = None,
        memcmp_opts: typing.Optional[typing.List[MemcmpOpts]] = None,
    ) -> typing.Sequence["AccountInfo"]:
        all_accounts = context.client.get_program_accounts(
            pubkey, data_slice=data_slice, data_size=data_size, memcmp_opts=memcmp_opts
        )

        all_account_infos = map(
            lambda result: AccountInfo._from_response_values(
                result["account"], PublicKey(result["pubkey"])
            ),
            all_accounts,
        )

        return list(all_account_infos)

    @staticmethod
    def _from_response_values(
        response_values: typing.Dict[str, typing.Any], address: PublicKey
    ) -> "AccountInfo":
        executable = bool(response_values["executable"])
        lamports = Decimal(response_values["lamports"])
        owner = PublicKey(response_values["owner"])
        rent_epoch = Decimal(response_values["rentEpoch"])
        data = decode_binary(response_values["data"])
        return AccountInfo(address, executable, lamports, owner, rent_epoch, data)

    @staticmethod
    def from_response(response: RPCResponse, address: PublicKey) -> "AccountInfo":
        return AccountInfo._from_response_values(response["result"]["value"], address)

    @staticmethod
    def _split_list_into_chunks(
        to_chunk: typing.Sequence[typing.Any], chunk_size: int = 100
    ) -> typing.Sequence[typing.Sequence[typing.Any]]:
        chunks = []
        start = 0
        while start < len(to_chunk):
            chunk = to_chunk[start : start + chunk_size]
            chunks += [chunk]
            start += chunk_size
        return chunks



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import argparse
import json
import logging
import os
import os.path
import typing

from solana.keypair import Keypair
from solana.publickey import PublicKey


# # 🥭 Wallet class
#
# The `Wallet` class wraps our understanding of saving and loading keys, and creating the
# appropriate Solana `Account` object.
#
# To load a private key from a file, the file must be a JSON-formatted text file with a root
# array of the 64 bytes making up the secret key.
#
# For example:
# ```
# [200,48,184,13... for another 60 bytes...]
# ```
#
# Alternatively (useful for some environments) the bytes can be loaded from the environment.
# The environment key is "KEYPAIR", so it would be stored in the environment using something
# like:
# ```
# export KEYPAIR="[200,48,184,13... for another 60 bytes...]"
# ```
# Alternatively, the environment key "SECRET_KEY" is accepted as an alias for KEYPAIR and may
# be used instead. (If both are specified, "KEYPAIR" is used.)
#
# **TODO:** It would be good to be able to load a `Wallet` from a mnemonic string. I haven't
# yet found a Python library that can generate a BIP44 derived seed for Solana that matches
# the derived seeds created by Sollet and Ledger.
#


_DEFAULT_WALLET_FILENAME: str = "id.json"


class Wallet:
    def __init__(self, secret_key: bytes) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.secret_key: bytes = secret_key[0:32]
        self.keypair: Keypair = Keypair.from_secret_key(self.secret_key)

    @property
    def address(self) -> PublicKey:
        return self.keypair.public_key

    def save(self, filename: str, overwrite: bool = False) -> None:
        if os.path.isfile(filename) and not overwrite:
            raise Exception(f"Wallet file '{filename}' already exists.")

        with open(filename, "w") as json_file:
            json.dump(list(self.keypair.secret_key), json_file)

    @staticmethod
    def load(filename: str = _DEFAULT_WALLET_FILENAME) -> "Wallet":
        if not os.path.isfile(filename):
            logging.error(f"Wallet file '{filename}' is not present.")
            raise Exception(f"Wallet file '{filename}' is not present.")
        else:
            with open(filename) as json_file:
                data = json.load(json_file)
                return Wallet(bytes(bytearray(data)))

    @staticmethod
    def create() -> "Wallet":
        new_account = Keypair.generate()
        return Wallet(new_account.secret_key)

    # Configuring a `Wallet` is a common operation for command-line programs and can involve a
    # lot of duplicate code.
    #
    # This function centralises some of it to ensure consistency and readability.
    #
    @staticmethod
    def add_command_line_parameters(parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--id-file",
            type=str,
            default=_DEFAULT_WALLET_FILENAME,
            help="file containing the JSON-formatted wallet private key",
        )

    # This function is the converse of `add_command_line_parameters()` - it takes
    # an argument of parsed command-line parameters and expects to see the ones it added
    # to that collection in the `add_command_line_parameters()` call.
    #
    # It then uses those parameters to create a properly-configured `Wallet` object.
    #
    @staticmethod
    def from_command_line_parameters(
        args: argparse.Namespace,
    ) -> typing.Optional["Wallet"]:
        # We always have an args.id_file (because we specify a default) so check for the environment
        # variable and give it priority.
        environment_secret_key = os.environ.get("KEYPAIR") or os.environ.get(
            "SECRET_KEY"
        )
        if environment_secret_key is not None:
            secret_key_bytes = json.loads(environment_secret_key)
            if len(secret_key_bytes) >= 32:
                return Wallet(bytes(secret_key_bytes))

        # Here we should have values for all our parameters.
        id_filename = args.id_file
        if os.path.isfile(id_filename):
            return Wallet.load(id_filename)

        return None

    @staticmethod
    def from_command_line_parameters_or_raise(args: argparse.Namespace) -> "Wallet":
        wallet = Wallet.from_command_line_parameters(args)
        if wallet is None:
            raise Exception("No wallet file or environment variables available.")

        return wallet

    def __str__(self) -> str:
        return f"« Wallet for {self.address} »"

    def __repr__(self) -> str:
        return f"{self}"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import argparse
import logging
import os
import typing

from decimal import Decimal
from solana.publickey import PublicKey
from solana.rpc.commitment import Commitment, Finalized

from .client import (
    AbstractSlotHolder,
    CheckingSlotHolder,
    ClusterUrlData,
    NullSlotHolder,
    TransactionMonitor,
    NullTransactionMonitor,
)
from .constants import EntropyConstants
from .context import Context
from .idsjsonmarketlookup import IdsJsonMarketLookup
from .instrumentlookup import (
    InstrumentLookup,
    CompoundInstrumentLookup,
    IdsJsonTokenLookup,
    NonSPLInstrumentLookup,
    SPLTokenLookup,
)
from .marketlookup import CompoundMarketLookup, MarketLookup
from .serummarketlookup import SerumMarketLookup
from .transactionmonitoring import (
    DequeTransactionStatusCollector,
    WebSocketTransactionMonitor,
)


# # 🥭 ContextBuilder
#
# ## Environment Variables
#
# It's possible to override the values in the `Context` variables provided. This can be easier than creating
# the `Context` in code or introducing dependencies and configuration.
#
# The following environment variables are read:
# * NAME
# * CLUSTER
# * CLUSTER_URL
# * GROUP_NAME
# * GROUP_ADDRESS
# * MANGO_PROGRAM_ADDRESS
# * SERUM_PROGRAM_ADDRESS
# * MANGO_REFLINK_ADDRESS


# # 🥭 ContextBuilder class
#
# A `ContextBuilder` class to allow building `Context` objects without introducing circular dependencies.
#
class ContextBuilder:

    # Utility class for parsing cluster-url command line parameter
    #
    class ParseClusterUrls(argparse.Action):
        cluster_urls: typing.List[ClusterUrlData] = []

        def __call__(
            self,
            parser: argparse.ArgumentParser,
            namespace: object,
            values: typing.Any,
            option_string: typing.Optional[str] = None,
        ) -> None:
            if values:
                if len(values) == 1:
                    self.cluster_urls.append(ClusterUrlData(rpc=values[0]))
                elif len(values) == 2:
                    self.cluster_urls.append(
                        ClusterUrlData(rpc=values[0], ws=values[1])
                    )
                else:
                    raise parser.error(
                        "Argument --cluster-url permits maximal two parameters. The first one configures HTTP connection url, the second one "
                        "configures the WS connection url. Example: --cluster-url https://localhost:8181 wss://localhost:8282"
                    )
                setattr(namespace, self.dest, self.cluster_urls)

    # Configuring a `Context` is a common operation for command-line programs and can involve a
    # lot of duplicate code.
    #
    # This function centralises some of it to ensure consistency and readability.
    #
    @staticmethod
    def add_command_line_parameters(
        parser: argparse.ArgumentParser, monitor_transactions_default: bool = False
    ) -> None:
        parser.add_argument(
            "--name",
            type=str,
            default="Entropy Explorer",
            help="Name of the program (used in reports and alerts)",
        )
        parser.add_argument(
            "--cluster-name", type=str, default=None, help="Solana RPC cluster name"
        )
        parser.add_argument(
            "--cluster-url",
            nargs="*",
            type=str,
            action=ContextBuilder.ParseClusterUrls,
            default=[],
            help="Solana RPC cluster URL (can be specified multiple times to provide failover when one errors; optional second parameter value defines websocket connection)",
        )
        parser.add_argument(
            "--group-name", type=str, default=None, help="Entropy group name"
        )
        parser.add_argument(
            "--group-address",
            type=PublicKey,
            default=None,
            help="Entropy group address",
        )
        parser.add_argument(
            "--entropy-program-address",
            type=PublicKey,
            default=None,
            help="Entropy program address",
        )
        parser.add_argument(
            "--serum-program-address",
            type=PublicKey,
            default=None,
            help="Serum program address",
        )
        parser.add_argument(
            "--skip-preflight",
            default=False,
            action="store_true",
            help="Skip pre-flight checks",
        )
        parser.add_argument(
            "--tpu-retransmissions",
            default=-1,
            type=int,
            help="Number of attempts the RPC node runs to deliver a transaction to TPU (default -1 means redelivering until success or recent blockhash expires)",
        )
        parser.add_argument(
            "--commitment",
            type=str,
            default=None,
            help="Commitment to use when sending transactions (can be 'finalized', 'confirmed' or 'processed')",
        )
        parser.add_argument(
            "--encoding",
            type=str,
            default=None,
            help="Encoding to request when receiving data from Solana (options are 'base58' (slow), 'base64', 'base64+zstd', or 'jsonParsed')",
        )
        parser.add_argument(
            "--blockhash-cache-duration",
            type=int,
            help="How long (in seconds) to cache 'recent' blockhashes",
        )
        parser.add_argument(
            "--http-request-timeout",
            type=float,
            default=20,
            help="What is the timeout for HTTP requests to when calling to RPC nodes (in seconds), -1 means no timeout",
        )
        parser.add_argument(
            "--stale-data-pause-before-retry",
            type=Decimal,
            help="How long (in seconds, e.g. 0.1) to pause after retrieving stale data before retrying",
        )
        parser.add_argument(
            "--stale-data-maximum-retries",
            type=int,
            default=0,
            help="How many times to retry fetching data after being given stale data before giving up",
        )
        parser.add_argument(
            "--monitor-transactions",
            default=monitor_transactions_default,
            action="store_true",
            help="Watch transactions and log their results",
        )
        parser.add_argument(
            "--monitor-transactions-commitment",
            type=Commitment,
            default=Finalized,
            help="Monitor transactions until they reach this commitment level",
        )
        parser.add_argument(
            "--monitor-transactions-timeout",
            type=float,
            default=90,
            help="Time after which to assume a transaction will not reach the specified commitment level",
        )
        parser.add_argument(
            "--gma-chunk-size",
            type=Decimal,
            default=None,
            help="Maximum number of addresses to send in a single call to getMultipleAccounts()",
        )
        parser.add_argument(
            "--gma-chunk-pause",
            type=Decimal,
            default=None,
            help="Number of seconds to pause between successive getMultipleAccounts() calls to avoid rate limiting",
        )
        parser.add_argument(
            "--reflink", type=PublicKey, default=None, help="Referral public key"
        )

    # This function is the converse of `add_command_line_parameters()` - it takes
    # an argument of parsed command-line parameters and expects to see the ones it added
    # to that collection in the `add_command_line_parameters()` call.
    #
    # It then uses those parameters to create a properly-configured `Context` object.
    #
    @staticmethod
    def from_command_line_parameters(args: argparse.Namespace) -> Context:
        name: typing.Optional[str] = args.name
        cluster_name: typing.Optional[str] = args.cluster_name
        cluster_urls: typing.Optional[
            typing.Sequence[ClusterUrlData]
        ] = args.cluster_url
        group_name: typing.Optional[str] = args.group_name
        group_address: typing.Optional[PublicKey] = args.group_address
        entropy_program_address: typing.Optional[
            PublicKey
        ] = args.entropy_program_address
        serum_program_address: typing.Optional[PublicKey] = args.serum_program_address
        skip_preflight: bool = bool(args.skip_preflight)
        tpu_retransmissions: int = int(args.tpu_retransmissions)
        commitment: typing.Optional[str] = args.commitment
        encoding: typing.Optional[str] = args.encoding
        blockhash_cache_duration: typing.Optional[int] = args.blockhash_cache_duration
        http_request_timeout: typing.Optional[float] = args.http_request_timeout
        stale_data_pause_before_retry: typing.Optional[
            Decimal
        ] = args.stale_data_pause_before_retry
        stale_data_maximum_retries: typing.Optional[
            int
        ] = args.stale_data_maximum_retries
        gma_chunk_size: typing.Optional[Decimal] = args.gma_chunk_size
        gma_chunk_pause: typing.Optional[Decimal] = args.gma_chunk_pause
        reflink: typing.Optional[PublicKey] = args.reflink
        monitor_transactions: bool = bool(args.monitor_transactions)
        monitor_transactions_commitment: typing.Optional[
            Commitment
        ] = args.monitor_transactions_commitment
        monitor_transactions_timeout: typing.Optional[
            float
        ] = args.monitor_transactions_timeout

        # Do this here so build() only ever has to handle the sequence of retry times. (It gets messy
        # passing around the sequnce *plus* the data to reconstruct it for build().)
        actual_stale_data_pauses_before_retry: typing.Sequence[float]
        actual_slot_holder: AbstractSlotHolder
        if stale_data_maximum_retries == 0:
            # Stale data checking is disabled
            actual_stale_data_pauses_before_retry = []
            actual_slot_holder = NullSlotHolder()
        else:
            retries: int = stale_data_maximum_retries or 10
            pause: Decimal = stale_data_pause_before_retry or Decimal("0.1")
            actual_stale_data_pauses_before_retry = [float(pause)] * retries
            actual_slot_holder = CheckingSlotHolder()

        context: Context = ContextBuilder.build(
            name,
            cluster_name,
            cluster_urls,
            skip_preflight,
            tpu_retransmissions,
            commitment,
            encoding,
            blockhash_cache_duration,
            http_request_timeout,
            actual_stale_data_pauses_before_retry,
            group_name,
            group_address,
            entropy_program_address,
            serum_program_address,
            gma_chunk_size,
            gma_chunk_pause,
            reflink,
            monitor_transactions,
            monitor_transactions_commitment,
            monitor_transactions_timeout,
            actual_slot_holder,
        )

        logging.debug(f"{context}")

        return context

    @staticmethod
    def default() -> Context:
        return ContextBuilder.build()

    @staticmethod
    def from_group_name(context: Context, group_name: str) -> Context:
        return ContextBuilder.build(
            context.name,
            context.client.cluster_name,
            context.client.cluster_urls,
            context.client.skip_preflight,
            context.client.tpu_retransmissions,
            context.client.commitment,
            context.client.encoding,
            context.client.blockhash_cache_duration,
            None,
            context.client.stale_data_pauses_before_retry,
            group_name,
            None,
            None,
            None,
            context.gma_chunk_size,
            context.gma_chunk_pause,
            context.reflink,
            not isinstance(context.client.transaction_monitor, NullTransactionMonitor),
            context.client.transaction_monitor.commitment,
            context.client.transaction_monitor.transaction_timeout,
            context.client.transaction_monitor.slot_holder,
        )

    @staticmethod
    def force_to_cluster(context: Context, cluster_name: str) -> Context:
        return ContextBuilder.build(
            context.name,
            cluster_name,
            None,  # Can't use parameter cluster URLs if we're switching network
            context.client.skip_preflight,
            context.client.tpu_retransmissions,
            context.client.commitment,
            context.client.encoding,
            context.client.blockhash_cache_duration,
            context.client.rpc_caller.all_providers[0].http_request_timeout,
            context.client.stale_data_pauses_before_retry,
            None,  # Use new cluster's default Group name
            None,  # Use new cluster's default Group address
            None,  # Use new cluster's default Program address
            None,  # Use new cluster's default Serum address
            context.gma_chunk_size,
            context.gma_chunk_pause,
            context.reflink,
            False,  # Don't try to watch transactions on a switched Context
            None,
            None,
            NullSlotHolder(),
        )

    @staticmethod
    def build(
        name: typing.Optional[str] = None,
        cluster_name: typing.Optional[str] = None,
        cluster_urls: typing.Optional[typing.Sequence[ClusterUrlData]] = None,
        skip_preflight: bool = False,
        tpu_retransmissions: int = -1,
        commitment: typing.Optional[str] = None,
        encoding: typing.Optional[str] = None,
        blockhash_cache_duration: typing.Optional[int] = None,
        http_request_timeout: typing.Optional[float] = None,
        stale_data_pauses_before_retry: typing.Optional[typing.Sequence[float]] = None,
        group_name: typing.Optional[str] = None,
        group_address: typing.Optional[PublicKey] = None,
        program_address: typing.Optional[PublicKey] = None,
        serum_program_address: typing.Optional[PublicKey] = None,
        gma_chunk_size: typing.Optional[Decimal] = None,
        gma_chunk_pause: typing.Optional[Decimal] = None,
        reflink: typing.Optional[PublicKey] = None,
        monitor_transactions: typing.Optional[bool] = None,
        monitor_transactions_commitment: typing.Optional[Commitment] = None,
        monitor_transactions_timeout: typing.Optional[float] = None,
        slot_holder: typing.Optional[AbstractSlotHolder] = None,
    ) -> "Context":
        def __public_key_or_none(
            address: typing.Optional[str],
        ) -> typing.Optional[PublicKey]:
            if address is not None and address != "":
                return PublicKey(address)
            return None

        # The first group is only used to determine the default cluster if it is not otherwise specified.
        first_group_data = EntropyConstants["groups"][0]
        actual_name: str = name or os.environ.get("NAME") or "Entropy Explorer"
        actual_cluster: str = (
            cluster_name
            or os.environ.get("CLUSTER_NAME")
            or first_group_data["cluster"]
        )

        # Now that we have the actual cluster name, taking environment variables and defaults into account,
        # we can decide what we want as the default group.
        for group_data in EntropyConstants["groups"]:
            if group_data["cluster"] == actual_cluster:
                default_group_data = group_data
                break

        actual_commitment: str = commitment or "processed"
        actual_encoding: str = encoding or "base64"
        actual_blockhash_cache_duration: int = blockhash_cache_duration or 0
        actual_stale_data_pauses_before_retry: typing.Sequence[float] = (
            stale_data_pauses_before_retry or []
        )
        actual_http_request_timeout: float = http_request_timeout or -1
        actual_tpu_retransmissions: int = int(tpu_retransmissions)

        actual_cluster_urls: typing.Optional[
            typing.Sequence[ClusterUrlData]
        ] = cluster_urls
        if actual_cluster_urls is None or len(actual_cluster_urls) == 0:
            cluster_url_from_environment: typing.Optional[str] = os.environ.get(
                "CLUSTER_URL"
            )
            if (
                cluster_url_from_environment is not None
                and cluster_url_from_environment != ""
            ):
                actual_cluster_urls = [ClusterUrlData(rpc=cluster_url_from_environment)]
            else:
                actual_cluster_urls = [
                    ClusterUrlData(rpc=EntropyConstants["cluster_urls"][actual_cluster])
                ]

        actual_skip_preflight: bool = skip_preflight
        actual_group_name: str = (
            group_name or os.environ.get("GROUP_NAME") or default_group_data["name"]
        )

        found_group_data: typing.Any = None
        for group in EntropyConstants["groups"]:
            if (
                group["cluster"] == actual_cluster
                and group["name"].upper() == actual_group_name.upper()
            ):
                found_group_data = group

        if found_group_data is None:
            raise Exception(
                f"Could not find group named '{actual_group_name}' in cluster '{actual_cluster}'."
            )
        # print("foundGroupData", found_group_data)
        actual_group_address: PublicKey = (
            group_address
            or __public_key_or_none(os.environ.get("GROUP_ADDRESS"))
            or PublicKey(found_group_data["publicKey"])
        )
        actual_program_address: PublicKey = (
            program_address
            or __public_key_or_none(os.environ.get("MANGO_PROGRAM_ADDRESS"))
            or PublicKey(found_group_data["entropyProgramId"])
        )
        actual_serum_program_address: PublicKey = (
            serum_program_address
            or __public_key_or_none(os.environ.get("SERUM_PROGRAM_ADDRESS"))
            or PublicKey(found_group_data["serumProgramId"])
        )

        actual_gma_chunk_size: Decimal = gma_chunk_size or Decimal(100)
        actual_gma_chunk_pause: Decimal = gma_chunk_pause or Decimal(0)

        actual_reflink: typing.Optional[PublicKey] = reflink or __public_key_or_none(
            os.environ.get("MANGO_REFLINK_ADDRESS")
        )

        ids_json_token_lookup: InstrumentLookup = IdsJsonTokenLookup(
            actual_cluster, actual_group_name
        )
        instrument_lookup: InstrumentLookup = ids_json_token_lookup
        if actual_cluster == "mainnet":
            # 'Overrides' are for problematic situations.
            #
            # We want to be able to use the community-owned SPL Token Registry JSON file. It holds details
            # of most tokens and allows our Serum code to work with any of them and their markets.
            #
            # The problems come when they decide to rename symbols, like they did with "ETH".
            #
            # Entropy uses "ETH" with a mint 2FPyTwcZLUg1MDrwsyoP4D6s1tM7hAkHYRjkNb5w6Pxk and for a long time
            # the SPL JSON file also used "ETH" with a mint of 2FPyTwcZLUg1MDrwsyoP4D6s1tM7hAkHYRjkNb5w6Pxk.
            #
            # Then the SPL JSON file was updated and the mint 2FPyTwcZLUg1MDrwsyoP4D6s1tM7hAkHYRjkNb5w6Pxk now
            # has the symbol 'soETH' for 'Wrapped Ethereum (Sollet)', and "ETH" is now 'Wrapped Ether (Wormhole)'
            # with a mint of 7vfCXTUXx5WJV5JADk17DUJ4ksgau7utNKj4b963voxs.
            #
            # 'Overrides' allows us to put the details we expect for 'ETH' into our loader, ahead of the SPL
            # JSON, so that our code and users can continue to use, for example, ETH/USDT, as they expect.
            mainnet_overrides_token_lookup: InstrumentLookup = SPLTokenLookup.load(
                SPLTokenLookup.OverridesDataFilepath
            )
            mainnet_spl_token_lookup: InstrumentLookup = SPLTokenLookup.load(
                SPLTokenLookup.DefaultDataFilepath
            )
            mainnet_non_spl_instrument_lookup: InstrumentLookup = (
                NonSPLInstrumentLookup.load(
                    NonSPLInstrumentLookup.DefaultMainnetDataFilepath
                )
            )
            instrument_lookup = CompoundInstrumentLookup(
                [
                    ids_json_token_lookup,
                    mainnet_overrides_token_lookup,
                    mainnet_non_spl_instrument_lookup,
                    mainnet_spl_token_lookup,
                ]
            )
        elif actual_cluster == "devnet":
            devnet_overrides_token_lookup: InstrumentLookup = SPLTokenLookup.load(
                SPLTokenLookup.DevnetOverridesDataFilepath
            )
            devnet_spl_token_lookup: InstrumentLookup = SPLTokenLookup.load(
                SPLTokenLookup.DevnetDataFilepath
            )
            devnet_non_spl_instrument_lookup: InstrumentLookup = (
                NonSPLInstrumentLookup.load(
                    NonSPLInstrumentLookup.DefaultDevnetDataFilepath
                )
            )
            instrument_lookup = CompoundInstrumentLookup(
                [
                    ids_json_token_lookup,
                    devnet_overrides_token_lookup,
                    devnet_non_spl_instrument_lookup,
                    devnet_spl_token_lookup,
                ]
            )

        ids_json_market_lookup: MarketLookup = IdsJsonMarketLookup(
            actual_cluster, instrument_lookup
        )
        all_market_lookup = ids_json_market_lookup
        if actual_cluster == "mainnet":
            mainnet_overrides_serum_market_lookup: SerumMarketLookup = (
                SerumMarketLookup.load(
                    actual_serum_program_address, SPLTokenLookup.OverridesDataFilepath
                )
            )
            mainnet_serum_market_lookup: SerumMarketLookup = SerumMarketLookup.load(
                actual_serum_program_address, SPLTokenLookup.DefaultDataFilepath
            )
            all_market_lookup = CompoundMarketLookup(
                [
                    ids_json_market_lookup,
                    mainnet_overrides_serum_market_lookup,
                    mainnet_serum_market_lookup,
                ]
            )
        elif actual_cluster == "devnet":
            devnet_overrides_serum_market_lookup: SerumMarketLookup = (
                SerumMarketLookup.load(
                    actual_serum_program_address,
                    SPLTokenLookup.DevnetOverridesDataFilepath,
                )
            )
            devnet_serum_market_lookup: SerumMarketLookup = SerumMarketLookup.load(
                actual_serum_program_address, SPLTokenLookup.DevnetDataFilepath
            )
            all_market_lookup = CompoundMarketLookup(
                [
                    ids_json_market_lookup,
                    devnet_overrides_serum_market_lookup,
                    devnet_serum_market_lookup,
                ]
            )
        market_lookup: MarketLookup = all_market_lookup

        actual_monitor_transactions_commitment: Commitment = (
            monitor_transactions_commitment or Finalized
        )
        actual_slot_holder: AbstractSlotHolder = slot_holder or NullSlotHolder()
        actual_monitor_transactions_timeout = monitor_transactions_timeout or 90
        actual_transaction_monitor: TransactionMonitor = NullTransactionMonitor(
            slot_holder=actual_slot_holder
        )
        if monitor_transactions:
            actual_transaction_monitor = WebSocketTransactionMonitor(
                actual_cluster_urls[0].ws,
                commitment=actual_monitor_transactions_commitment,
                transaction_timeout=actual_monitor_transactions_timeout,
                collector=DequeTransactionStatusCollector(),
                slot_holder=actual_slot_holder,
            )

        context = Context(
            actual_name,
            actual_cluster,
            actual_cluster_urls,
            actual_skip_preflight,
            actual_tpu_retransmissions,
            actual_commitment,
            actual_encoding,
            actual_blockhash_cache_duration,
            actual_http_request_timeout,
            actual_stale_data_pauses_before_retry,
            actual_program_address,
            actual_serum_program_address,
            actual_group_name,
            actual_group_address,
            actual_gma_chunk_size,
            actual_gma_chunk_pause,
            actual_reflink,
            instrument_lookup,
            market_lookup,
            actual_transaction_monitor,
        )

        return context



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import logging
import typing

from .version import Version


# # 🥭 AccountFlags class
#
# Encapsulates the Serum AccountFlags data.
#


class AccountFlags:
    def __init__(
        self,
        version: Version,
        initialized: bool,
        market: bool,
        open_orders: bool,
        request_queue: bool,
        event_queue: bool,
        bids: bool,
        asks: bool,
        disabled: bool,
    ) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.version: Version = version
        self.initialized: bool = initialized
        self.market: bool = market
        self.open_orders: bool = open_orders
        self.request_queue: bool = request_queue
        self.event_queue: bool = event_queue
        self.bids: bool = bids
        self.asks: bool = asks
        self.disabled: bool = disabled

    @staticmethod
    def from_layout(layout: typing.Any) -> "AccountFlags":
        return AccountFlags(
            Version.UNSPECIFIED,
            layout.initialized,
            layout.market,
            layout.open_orders,
            layout.request_queue,
            layout.event_queue,
            layout.bids,
            layout.asks,
            layout.disabled,
        )

    def __str__(self) -> str:
        flags: typing.List[typing.Optional[str]] = []
        flags += ["initialized" if self.initialized else None]
        flags += ["market" if self.market else None]
        flags += ["open_orders" if self.open_orders else None]
        flags += ["request_queue" if self.request_queue else None]
        flags += ["event_queue" if self.event_queue else None]
        flags += ["bids" if self.bids else None]
        flags += ["asks" if self.asks else None]
        flags += ["disabled" if self.disabled else None]
        flag_text = " | ".join(flag for flag in flags if flag is not None) or "None"
        return f"« AccountFlags: {flag_text} »"

    def __repr__(self) -> str:
        return f"{self}"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import abc
import typing

from datetime import datetime
from decimal import Decimal
from solana.publickey import PublicKey

from .accountinfo import AccountInfo
from .addressableaccount import AddressableAccount
from .context import Context
from .layouts import layouts
from .lotsizeconverter import LotSizeConverter
from .metadata import Metadata
from .observables import Disposable
from .orders import Side
from .version import Version
from .websocketsubscription import (
    WebSocketAccountSubscription,
    WebSocketSubscriptionManager,
)


# # 🥭 PerpEvent class
#
# `PerpEvent` is the base class of all perp event objects.
#
class PerpEvent(metaclass=abc.ABCMeta):
    def __init__(self, event_type: int, original_index: Decimal) -> None:
        self.event_type: int = event_type
        self.original_index: Decimal = original_index

    @abc.abstractproperty
    @property
    def accounts_to_crank(self) -> typing.Sequence[PublicKey]:
        raise NotImplementedError(
            "PerpEvent.accounts_to_crank is not implemented on the base type."
        )

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 PerpFillEvent class
#
# `PerpOutEvent` stores details of a perp 'fill' event.
#
class PerpFillEvent(PerpEvent):
    def __init__(
        self,
        event_type: int,
        original_index: Decimal,
        timestamp: datetime,
        taker_side: Side,
        price: Decimal,
        quantity: Decimal,
        best_initial: Decimal,
        maker_slot: Decimal,
        maker_out: bool,
        maker: PublicKey,
        maker_order_id: Decimal,
        maker_client_order_id: Decimal,
        taker: PublicKey,
        taker_order_id: Decimal,
        taker_client_order_id: Decimal,
    ) -> None:
        super().__init__(event_type, original_index)
        self.timestamp: datetime = timestamp
        self.taker_side: Side = taker_side
        self.price: Decimal = price
        self.quantity: Decimal = quantity

        self.best_initial: Decimal = best_initial
        self.maker_slot: Decimal = maker_slot
        self.maker_out: bool = maker_out

        self.maker: PublicKey = maker
        self.maker_order_id: Decimal = maker_order_id
        self.maker_client_order_id: Decimal = maker_client_order_id

        self.taker: PublicKey = taker
        self.taker_order_id: Decimal = taker_order_id
        self.taker_client_order_id: Decimal = taker_client_order_id

    @property
    def key(self) -> str:
        return f"{self.maker_order_id}/{self.taker_order_id}"

    @property
    def accounts_to_crank(self) -> typing.Sequence[PublicKey]:
        return [self.maker, self.taker]

    def __str__(self) -> str:
        return f"""« PerpFillEvent [{self.original_index}] {self.taker_side} {self.quantity:,.8f} at {self.price:,.8f} [{self.timestamp}]
    Maker: {self.maker}, {self.maker_order_id} / {self.maker_client_order_id}
    Taker: {self.taker}, {self.taker_order_id} / {self.taker_client_order_id}
    Best Initial: {self.best_initial}
    Maker Slot: {self.maker_slot}
    Maker Out: {self.maker_out}
»"""


# # 🥭 PerpOutEvent class
#
# `PerpOutEvent` stores details of a perp 'out' event.
#
class PerpOutEvent(PerpEvent):
    def __init__(
        self,
        event_type: int,
        original_index: Decimal,
        owner: PublicKey,
        side: Side,
        quantity: Decimal,
        slot: Decimal,
    ) -> None:
        super().__init__(event_type, original_index)
        self.owner: PublicKey = owner
        self.side: Side = side
        self.slot: Decimal = slot
        self.quantity: Decimal = quantity

    @property
    def accounts_to_crank(self) -> typing.Sequence[PublicKey]:
        return [self.owner]

    def __str__(self) -> str:
        return f"""« PerpOutEvent [{self.original_index}] [{self.owner}] {self.side} {self.quantity}, slot: {self.slot} »"""


# # 🥭 PerpLiquidateEvent class
#
# `PerpLiquidateEvent` stores details of a perp 'liquidate' event.
#
class PerpLiquidateEvent(PerpEvent):
    def __init__(
        self,
        event_type: int,
        original_index: Decimal,
        timestamp: datetime,
        seq_num: Decimal,
        liquidatee: PublicKey,
        liquidator: PublicKey,
        price: Decimal,
        quantity: Decimal,
        liquidation_fee: Decimal,
    ) -> None:
        super().__init__(event_type, original_index)
        self.timestamp: datetime = timestamp
        self.seq_num: Decimal = seq_num
        self.liquidatee: PublicKey = liquidatee
        self.liquidator: PublicKey = liquidator
        self.price: Decimal = price
        self.quantity: Decimal = quantity
        self.liquidation_fee: Decimal = liquidation_fee

    @property
    def accounts_to_crank(self) -> typing.Sequence[PublicKey]:
        return [self.liquidatee, self.liquidator]

    def __str__(self) -> str:
        return f"""« PerpLiquidateEvent Liquidator {self.liquidator} liquidated {self.liquidatee} with {self.quantity} at {self.price} »"""


# # 🥭 PerpUnknownEvent class
#
# `PerpUnknownEvent` details an unknown `PerpEvent`. This should never be encountered, but might if
# the event queue data is upgraded before this code.
#
class PerpUnknownEvent(PerpEvent):
    def __init__(
        self, event_type: int, original_index: Decimal, owner: PublicKey
    ) -> None:
        super().__init__(event_type, original_index)
        self.owner: PublicKey = owner

    @property
    def accounts_to_crank(self) -> typing.Sequence[PublicKey]:
        return [self.owner]

    def __str__(self) -> str:
        return f"« PerpUnknownEvent [{self.original_index}] [{self.owner}] »"


# # 🥭 event_builder function
#
# `event_builder()` takes an event layout and returns a typed `PerpEvent`.
#
def event_builder(
    lot_size_converter: LotSizeConverter,
    event_layout: typing.Any,
    original_index: Decimal,
) -> typing.Optional[PerpEvent]:
    if event_layout.event_type == b"\x00":
        if event_layout.maker is None and event_layout.taker is None:
            return None
        taker_side: Side = Side.from_value(event_layout.taker_side)
        quantity: Decimal = lot_size_converter.base_size_lots_to_number(
            event_layout.quantity
        )
        price: Decimal = lot_size_converter.price_lots_to_number(event_layout.price)
        return PerpFillEvent(
            event_layout.event_type,
            original_index,
            event_layout.timestamp,
            taker_side,
            price,
            quantity,
            event_layout.best_initial,
            event_layout.maker_slot,
            event_layout.maker_out,
            event_layout.maker,
            event_layout.maker_order_id,
            event_layout.maker_client_order_id,
            event_layout.taker,
            event_layout.taker_order_id,
            event_layout.taker_client_order_id,
        )
    elif event_layout.event_type == b"\x01":
        return PerpOutEvent(
            event_layout.event_type,
            original_index,
            event_layout.owner,
            event_layout.side,
            event_layout.quantity,
            event_layout.slot,
        )
    elif event_layout.event_type == b"\x02":
        return PerpLiquidateEvent(
            event_layout.event_type,
            original_index,
            event_layout.timestamp,
            event_layout.seq_num,
            event_layout.liquidatee,
            event_layout.liquidator,
            event_layout.price,
            event_layout.quantity,
            event_layout.liquidation_fee,
        )
    else:
        return PerpUnknownEvent(
            event_layout.event_type, original_index, event_layout.owner
        )


# # 🥭 PerpEventQueue class
#
# `PerpEventQueue` stores details of perp events in a ringbuffer, along with indices to track which events are
# processed by 'consume events' and which are not.
#
class PerpEventQueue(AddressableAccount):
    def __init__(
        self,
        account_info: AccountInfo,
        version: Version,
        meta_data: Metadata,
        lot_size_converter: LotSizeConverter,
        head: Decimal,
        count: Decimal,
        sequence_number: Decimal,
        unprocessed_events: typing.Sequence[PerpEvent],
        processed_events: typing.Sequence[PerpEvent],
    ) -> None:
        super().__init__(account_info)
        self.version: Version = version

        self.meta_data: Metadata = meta_data
        self.lot_size_converter: LotSizeConverter = lot_size_converter
        self.head: Decimal = head
        self.count: Decimal = count
        self.sequence_number: Decimal = sequence_number

        self.unprocessed_events: typing.Sequence[PerpEvent] = unprocessed_events
        self.processed_events: typing.Sequence[PerpEvent] = processed_events

    @staticmethod
    def from_layout(
        layout: typing.Any,
        account_info: AccountInfo,
        version: Version,
        lot_size_converter: LotSizeConverter,
    ) -> "PerpEventQueue":
        meta_data: Metadata = Metadata.from_layout(layout.meta_data)
        head: Decimal = layout.head
        count: Decimal = layout.count
        seq_num: Decimal = layout.seq_num
        events: typing.List[PerpEvent] = []
        for index, raw_event in enumerate(layout.events):
            built_event = event_builder(lot_size_converter, raw_event, Decimal(index))
            if built_event is not None:
                events += [built_event]

        # Events are stored in a ringbuffer, and the oldest is overwritten when a new event arrives.
        # Make it a bit simpler to use by splitting at the insertion point and swapping the two pieces
        # around so that users don't have to do modulo arithmetic on the capacity.
        ordered_events = events[int(head) :] + events[0 : int(head)]

        # Now chop the oldest-to-newest list of events into processed and unprocessed. The `count`
        # property holds the number of unprocessed events.
        unprocessed_events = ordered_events[0 : int(count)]
        processed_events = ordered_events[int(count) :]

        return PerpEventQueue(
            account_info,
            version,
            meta_data,
            lot_size_converter,
            head,
            count,
            seq_num,
            unprocessed_events,
            processed_events,
        )

    @staticmethod
    def parse(
        account_info: AccountInfo, lot_size_converter: LotSizeConverter
    ) -> "PerpEventQueue":
        # Data length isn't fixed so can't check we get the right value the way we normally do.
        layout = layouts.PERP_EVENT_QUEUE.parse(account_info.data)
        return PerpEventQueue.from_layout(
            layout, account_info, Version.V1, lot_size_converter
        )

    @staticmethod
    def load(
        context: Context, address: PublicKey, lot_size_converter: LotSizeConverter
    ) -> "PerpEventQueue":
        account_info = AccountInfo.load(context, address)
        if account_info is None:
            raise Exception(f"PerpEventQueue account not found at address '{address}'")
        return PerpEventQueue.parse(account_info, lot_size_converter)

    @property
    def accounts_to_crank(self) -> typing.Sequence[PublicKey]:
        to_crank: typing.List[PublicKey] = []
        for event_to_crank in self.unprocessed_events:
            to_crank += event_to_crank.accounts_to_crank

        seen = []
        distinct = []
        for account in to_crank:
            account_str = account.to_base58()
            if account_str not in seen:
                distinct += [account]
                seen += [account_str]

        return distinct

    @property
    def events(self) -> typing.Sequence[PerpEvent]:
        return [*self.processed_events, *self.unprocessed_events]

    @property
    def fills(self) -> typing.Sequence[PerpFillEvent]:
        fills: typing.List[PerpFillEvent] = []
        for event in self.events:
            if isinstance(event, PerpFillEvent):
                fills += [event]
        return fills

    @property
    def liquidations(self) -> typing.Sequence[PerpLiquidateEvent]:
        liquidations: typing.List[PerpLiquidateEvent] = []
        for event in self.events:
            if isinstance(event, PerpLiquidateEvent):
                liquidations += [event]
        return liquidations

    def subscribe(
        self,
        context: Context,
        websocketmanager: WebSocketSubscriptionManager,
        callback: typing.Callable[["PerpEventQueue"], None],
    ) -> Disposable:
        def __parser(account_info: AccountInfo) -> PerpEventQueue:
            return PerpEventQueue.parse(account_info, self.lot_size_converter)

        subscription = WebSocketAccountSubscription(context, self.address, __parser)
        websocketmanager.add(subscription)
        subscription.publisher.subscribe(on_next=callback)  # type: ignore[call-arg]

        return subscription

    def events_for_account(
        self, entropy_account_address: PublicKey
    ) -> typing.Sequence[PerpEvent]:
        events: typing.List[PerpEvent] = []
        for event in self.events:
            if entropy_account_address in event.accounts_to_crank:
                events += [event]
        return events

    def fills_for_account(
        self, entropy_account_address: PublicKey
    ) -> typing.Sequence[PerpFillEvent]:
        fills: typing.List[PerpFillEvent] = []
        for event in self.events_for_account(entropy_account_address):
            if isinstance(event, PerpFillEvent):
                fills += [event]
        return fills

    @property
    def capacity(self) -> int:
        return len(self.unprocessed_events) + len(self.processed_events)

    def __str__(self) -> str:
        unprocessed_events = (
            "\n        ".join(
                [
                    f"{event}".replace("\n", "\n        ")
                    for event in self.unprocessed_events
                    if event is not None
                ]
            )
            or "None"
        )
        processed_events = (
            "\n        ".join(
                [
                    f"{event}".replace("\n", "\n        ")
                    for event in self.processed_events
                    if event is not None
                ]
            )
            or "None"
        )
        return f"""« PerpEventQueue [{self.version}] {self.address}
    {self.meta_data}
    Head: {self.head}
    Count: {self.count}
    Sequence Number: {self.sequence_number}
    Capacity: {self.capacity}
    Unprocessed Events:
        {unprocessed_events}
    Processed Events:
        {processed_events}
»"""


# # 🥭 UnseenPerpEventChangesTracker class
#
# `UnseenPerpEventChangesTracker` tracks changes to a specific `PerpEventQueue`. When an updated version
# of the `PerpEventQueue` is passed to `unseen()`, any new events are returned.
#
# Seen events are tracked by keeping a 'sequence_number' index of the last event seen in the
# `PerpEventQueue` ringbuffer. When a new `PerpEventQueue` appears, the difference between its
# sequence_number and the stored sequence_number is used to calculate the number of new events
# to return.
#
class UnseenPerpEventChangesTracker:
    def __init__(self, initial: PerpEventQueue) -> None:
        self.last_sequence_number: Decimal = initial.sequence_number

    def unseen(self, event_queue: PerpEventQueue) -> typing.Sequence[PerpEvent]:
        unseen: typing.List[PerpEvent] = []
        new_sequence_number: Decimal = event_queue.sequence_number
        if self.last_sequence_number != new_sequence_number:
            number_of_changes: Decimal = new_sequence_number - self.last_sequence_number
            unseen = [*event_queue.processed_events, *event_queue.unprocessed_events][
                0 - int(number_of_changes) :
            ]
            self.last_sequence_number = new_sequence_number

        return unseen


# # 🥭 UnseenAccountFillEventTracker class
#
# `UnseenAccountFillEventTracker` tracks fills for a specific Entropy Account in a specific
# `PerpEventQueue`. When an updated version of the `PerpEventQueue` is passed to `unseen()`,
# any new fills are returned.
#
class UnseenAccountFillEventTracker:
    def __init__(
        self, initial: PerpEventQueue, entropy_account_address: PublicKey
    ) -> None:
        self.entropy_account_address: PublicKey = entropy_account_address
        self.last_sequence_number: Decimal = initial.sequence_number
        initial_fills: typing.Sequence[PerpFillEvent] = initial.fills_for_account(
            entropy_account_address
        )
        self.last_key: str = initial_fills[-1].key if len(initial_fills) > 0 else ""

    def unseen(self, event_queue: PerpEventQueue) -> typing.Sequence[PerpEvent]:
        fills: typing.Sequence[PerpFillEvent] = event_queue.fills_for_account(
            self.entropy_account_address
        )
        if len(fills) == 0:
            return []

        last_key_position: int = -1
        for counter, fill in enumerate(fills):
            if fill.key == self.last_key:
                last_key_position = counter
                break

        unseen: typing.Sequence[PerpFillEvent]
        if last_key_position == -1:
            # We haven't seen any of these fills
            unseen = fills
        elif last_key_position == len(fills) - 1:
            unseen = []
        else:
            unseen = fills[last_key_position + 1 :]
            self.last_key = unseen[-1].key

        return unseen



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import abc
import logging
import typing

from datetime import datetime
from decimal import Decimal
from solana.publickey import PublicKey

from .combinableinstructions import CombinableInstructions
from .constants import SYSTEM_PROGRAM_ADDRESS
from .datetimes import utc_now
from .loadedmarket import LoadedMarket
from .lotsizeconverter import NullLotSizeConverter
from .markets import InventorySource
from .orders import Order, OrderBook, OrderType, Side


# # 🥭 MarketOperations
#
# This file deals with placing orders. We want the interface to be simple and basic:
# ```
# market_operations.cancel_order(old_order)
# market_operations.place_order(new_order)
# ```
# This requires the `MarketOperations` already know a bit about the market it is placing the
# order on, and the code in the `MarketOperations` be specialised for that market platform.
#


# # 🥭 MarketInstructionBuilder class
#
# This abstracts the process of buiding instructions for placing orders and cancelling orders.
#
# It's abstracted because we may want to have different implementations for different market types.
#
# Whichever choice is made, the calling code shouldn't have to care. It should be able to
# use its `MarketInstructionBuilder` class as simply as:
# ```
# instruction_builder.build_cancel_order_instructions(order)
# ```
#
# As a matter of policy for all InstructionBuidlers, construction and build_* methods should all work with
# existing data, requiring no fetches from Solana or other sources. All necessary data should all be loaded
# on initial setup in the `load()` method.
#
class MarketInstructionBuilder(metaclass=abc.ABCMeta):
    def __init__(self) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.open_orders_address: typing.Optional[PublicKey] = None

    @abc.abstractmethod
    def build_cancel_order_instructions(
        self, order: Order, ok_if_missing: bool = False
    ) -> CombinableInstructions:
        raise NotImplementedError(
            "MarketInstructionBuilder.build_cancel_order_instructions() is not implemented on the base type."
        )

    @abc.abstractmethod
    def build_place_order_instructions(self, order: Order) -> CombinableInstructions:
        raise NotImplementedError(
            "MarketInstructionBuilder.build_place_order_instructions() is not implemented on the base type."
        )

    @abc.abstractmethod
    def build_settle_instructions(self) -> CombinableInstructions:
        raise NotImplementedError(
            "MarketInstructionBuilder.build_settle_instructions() is not implemented on the base type."
        )

    @abc.abstractmethod
    def build_crank_instructions(
        self, addresses: typing.Sequence[PublicKey], limit: Decimal = Decimal(32)
    ) -> CombinableInstructions:
        raise NotImplementedError(
            "MarketInstructionBuilder.build_crank_instructions() is not implemented on the base type."
        )

    @abc.abstractmethod
    def build_redeem_instructions(self) -> CombinableInstructions:
        raise NotImplementedError(
            "MarketInstructionBuilder.build_redeem_instructions() is not implemented on the base type."
        )

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 MarketOperations class
#
# This abstracts the process of placing orders, providing a base class for specialised operations.
#
# It's abstracted because we may want to have different approaches to placing these
# orders - do we want to run them against the Serum orderbook? Do we want to run them against
# Entropy groups?
#
# Whichever choice is made, the calling code shouldn't have to care. It should be able to
# use its `MarketOperations` class as simply as:
# ```
# market_operations.place_order(order)
# ```
#
class MarketOperations(metaclass=abc.ABCMeta):
    def __init__(self, market: LoadedMarket) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.market: LoadedMarket = market

    @property
    def symbol(self) -> str:
        return self.market.fully_qualified_symbol

    @property
    def inventory_source(self) -> InventorySource:
        return self.market.inventory_source

    @abc.abstractmethod
    def cancel_order(
        self, order: Order, ok_if_missing: bool = False
    ) -> typing.Sequence[str]:
        raise NotImplementedError(
            "MarketOperations.cancel_order() is not implemented on the base type."
        )

    @abc.abstractmethod
    def place_order(
        self, order: Order, crank_limit: Decimal = Decimal(5)
    ) -> typing.Sequence[str]:
        raise NotImplementedError(
            "MarketOperations.place_order() is not implemented on the base type."
        )

    @abc.abstractmethod
    def load_orderbook(self) -> OrderBook:
        raise NotImplementedError(
            "MarketOperations.load_orders() is not implemented on the base type."
        )

    @abc.abstractmethod
    def load_my_orders(
        self, cutoff: typing.Optional[datetime] = utc_now()
    ) -> typing.Sequence[Order]:
        raise NotImplementedError(
            "MarketOperations.load_my_orders() is not implemented on the base type."
        )

    @abc.abstractmethod
    def settle(self) -> typing.Sequence[str]:
        raise NotImplementedError(
            "MarketOperations.settle() is not implemented on the base type."
        )

    @abc.abstractmethod
    def crank(self, limit: Decimal = Decimal(32)) -> typing.Sequence[str]:
        raise NotImplementedError(
            "MarketOperations.crank() is not implemented on the base type."
        )

    @abc.abstractmethod
    def create_openorders(self) -> PublicKey:
        raise NotImplementedError(
            "MarketOperations.create_openorders() is not implemented on the base type."
        )

    @abc.abstractmethod
    def ensure_openorders(self) -> PublicKey:
        raise NotImplementedError(
            "MarketOperations.ensure_openorders() is not implemented on the base type."
        )

    def market_buy(
        self, quantity: Decimal, max_slippage: Decimal
    ) -> typing.Sequence[str]:
        orderbook = self.load_orderbook()
        if orderbook.top_ask is None:
            raise Exception(f"Could not determine top ask on {orderbook.symbol}")

        top_ask = orderbook.top_ask.price

        increase_factor = Decimal(1) + max_slippage
        price = top_ask * increase_factor
        self._logger.info(f"Price {price} - adjusted by {max_slippage} from {top_ask}")

        order = Order.from_values(Side.BUY, price, quantity, OrderType.IOC)
        return self.place_order(order)

    def market_sell(
        self, quantity: Decimal, max_slippage: Decimal
    ) -> typing.Sequence[str]:
        orderbook = self.load_orderbook()
        if orderbook.top_bid is None:
            raise Exception(f"Could not determine top bid on {orderbook.symbol}")

        top_bid = orderbook.top_bid.price

        decrease_factor = Decimal(1) - max_slippage
        price = top_bid * decrease_factor
        self._logger.info(f"Price {price} - adjusted by {max_slippage} from {top_bid}")

        order = Order.from_values(Side.SELL, price, quantity, OrderType.IOC)
        return self.place_order(order)

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 NullMarketInstructionBuilder class
#
# A null, no-op, dry-run trade executor that can be plugged in anywhere a `MarketInstructionBuilder`
# is expected, but which will not actually trade.
#
class NullMarketInstructionBuilder(MarketInstructionBuilder):
    def __init__(self, symbol: str) -> None:
        super().__init__()
        self.symbol: str = symbol

    def build_cancel_order_instructions(
        self, order: Order, ok_if_missing: bool = False
    ) -> CombinableInstructions:
        return CombinableInstructions.empty()

    def build_place_order_instructions(self, order: Order) -> CombinableInstructions:
        return CombinableInstructions.empty()

    def build_settle_instructions(self) -> CombinableInstructions:
        return CombinableInstructions.empty()

    def build_crank_instructions(
        self, addresses: typing.Sequence[PublicKey], limit: Decimal = Decimal(32)
    ) -> CombinableInstructions:
        return CombinableInstructions.empty()

    def build_redeem_instructions(self) -> CombinableInstructions:
        return CombinableInstructions.empty()

    def __str__(self) -> str:
        return f"« NullMarketInstructionBuilder {self.symbol} »"


# # 🥭 NullMarketOperations class
#
# A null, no-op, dry-run trade executor that can be plugged in anywhere a `MarketOperations`
# is expected, but which will not actually trade.
#
class NullMarketOperations(MarketOperations):
    def __init__(self, market: LoadedMarket) -> None:
        super().__init__(market)

    def cancel_order(
        self, order: Order, ok_if_missing: bool = False
    ) -> typing.Sequence[str]:
        self._logger.info(f"[Dry Run] Not cancelling order {order}.")
        return [""]

    def place_order(
        self, order: Order, crank_limit: Decimal = Decimal(5)
    ) -> typing.Sequence[str]:
        self._logger.info(f"[Dry Run] Not placing order {order}.")
        return []

    def load_orderbook(self) -> OrderBook:
        return OrderBook(
            self.market.fully_qualified_symbol, NullLotSizeConverter(), [], []
        )

    def load_my_orders(
        self, cutoff: typing.Optional[datetime] = utc_now()
    ) -> typing.Sequence[Order]:
        return []

    def settle(self) -> typing.Sequence[str]:
        return []

    def crank(self, limit: Decimal = Decimal(32)) -> typing.Sequence[str]:
        return []

    def create_openorders(self) -> PublicKey:
        return SYSTEM_PROGRAM_ADDRESS

    def ensure_openorders(self) -> PublicKey:
        return SYSTEM_PROGRAM_ADDRESS

    def market_buy(
        self, quantity: Decimal, max_slippage: Decimal
    ) -> typing.Sequence[str]:
        self._logger.info(
            f"[Dry Run] Not market buying {quantity} with max slippage {max_slippage}."
        )
        return []

    def market_sell(
        self, quantity: Decimal, max_slippage: Decimal
    ) -> typing.Sequence[str]:
        self._logger.info(
            f"[Dry Run] Not market selling {quantity} with max slippage {max_slippage}."
        )
        return []

    def __str__(self) -> str:
        return f"""« NullMarketOperations [{self.market.fully_qualified_symbol}] »"""



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import logging
import typing

from decimal import Decimal

from .account import Account
from .cache import Cache
from .group import Group
from .instrumentvalue import InstrumentValue
from .loadedmarket import LoadedMarket
from .markets import InventorySource
from .openorders import OpenOrders
from .perpmarket import PerpMarket
from .spotmarket import SpotMarket
from .watcher import Watcher


# # 🥭 Inventory class
#
# This class details inventory of a crypto account for a market.
#
class Inventory:
    def __init__(
        self,
        inventory_source: InventorySource,
        liquidity_incentives: InstrumentValue,
        available_collateral: InstrumentValue,
        base: InstrumentValue,
        quote: InstrumentValue,
    ) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.inventory_source: InventorySource = inventory_source
        self.available_collateral: InstrumentValue = available_collateral
        self.liquidity_incentives: InstrumentValue = liquidity_incentives
        self.base: InstrumentValue = base
        self.quote: InstrumentValue = quote

    @property
    def symbol(self) -> str:
        return f"{self.base.token.symbol}/{self.quote.token.symbol}"

    def __str__(self) -> str:
        liquidity_incentives: str = ""
        if self.liquidity_incentives.value > 0:
            liquidity_incentives = f" {self.liquidity_incentives}"
        return f"« Inventory {self.symbol}{liquidity_incentives} [{self.base} / {self.quote}] ({self.available_collateral} available) »"

    def __repr__(self) -> str:
        return f"{self}"


class InventoryAccountWatcher:
    def __init__(
        self,
        market: LoadedMarket,
        account_watcher: Watcher[Account],
        group_watcher: Watcher[Group],
        all_open_orders_watchers: typing.Sequence[Watcher[OpenOrders]],
        cache_watcher: Watcher[Cache],
    ):
        self.account_watcher: Watcher[Account] = account_watcher
        self.group_watcher: Watcher[Group] = group_watcher
        self.all_open_orders_watchers: typing.Sequence[
            Watcher[OpenOrders]
        ] = all_open_orders_watchers
        self.cache_watcher: Watcher[Cache] = cache_watcher
        account: Account = account_watcher.latest
        if SpotMarket.isa(market):
            self.spot_account_index: int = (
                group_watcher.latest.slot_by_spot_market_address(market.address).index
            )
        elif PerpMarket.isa(market):
            self.spot_account_index = group_watcher.latest.slot_by_perp_market_address(
                market.address
            ).index
        else:
            raise Exception(
                f"Cannot find slot for market {market} in group {group_watcher.latest.address}"
            )

        base_value = InstrumentValue.find_by_symbol(
            account.net_values, market.base.symbol
        )
        self.base_index: int = account.net_values_by_index.index(base_value)
        quote_value = InstrumentValue.find_by_symbol(
            account.net_values, market.quote.symbol
        )
        self.quote_index: int = account.net_values_by_index.index(quote_value)

    @property
    def latest(self) -> Inventory:
        account: Account = self.account_watcher.latest
        group: Group = self.group_watcher.latest
        cache: Cache = self.cache_watcher.latest

        # Spot markets don't accrue MNGO liquidity incentives
        mngo = group.liquidity_incentive_token
        mngo_accrued: InstrumentValue = InstrumentValue(mngo, Decimal(0))

        all_open_orders: typing.Dict[str, OpenOrders] = {
            str(oo_watcher.latest.address): oo_watcher.latest
            for oo_watcher in self.all_open_orders_watchers
        }

        frame = account.to_dataframe(group, all_open_orders, cache)
        available_collateral: InstrumentValue = account.init_health(frame)

        base_value = account.net_values_by_index[self.base_index]
        if base_value is None:
            raise Exception(
                f"Could not find net assets in account {account.address} at index {self.base_index}."
            )
        quote_value = account.net_values_by_index[self.quote_index]
        if quote_value is None:
            raise Exception(
                f"Could not find net assets in account {account.address} at index {self.quote_index}."
            )

        return Inventory(
            InventorySource.ACCOUNT,
            mngo_accrued,
            available_collateral,
            base_value,
            quote_value,
        )



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import decimal
import importlib.metadata
import json
import os.path
import typing

from dataclasses import dataclass
from solana.publickey import PublicKey


# # 🥭 Constants
#
# This file contains some hard-coded values, all kept in one place, as well as the mechanism
# for loading the Entropy `ids.json` file.

# ## SYSTEM_PROGRAM_ADDRESS
#
# The Solana system program address is always 11111111111111111111111111111111.
#
SYSTEM_PROGRAM_ADDRESS = PublicKey("11111111111111111111111111111111")


# ## SOL_MINT_ADDRESS
#
# The fake mint address of the SOL token. **Note:** Wrapped SOL has a different mint address - it is So11111111111111111111111111111111111111112.
#
SOL_MINT_ADDRESS = PublicKey("So11111111111111111111111111111111111111111")


# ## SOL_DECIMALS
#
# The number of decimal places used to convert Lamports into SOLs.
#
SOL_DECIMALS = decimal.Decimal(9)


# ## SOL_DECIMAL_DIVISOR decimal
#
# The divisor to use to turn an integer value of SOLs from an account's `balance` into a value with the correct number of decimal places.
#
SOL_DECIMAL_DIVISOR = decimal.Decimal(10**SOL_DECIMALS)


# ## I64_MAX
#
# This is to match the entropy-client-v3 behaviour:
# export const I64_MAX_BN = new BN('9223372036854775807').toTwos(64);
I64_MAX = decimal.Decimal("9223372036854775807")


# # WARNING_DISCLAIMER_TEXT
#
# This is the warning text that is output on each run of a command.
#
WARNING_DISCLAIMER_TEXT = """
⚠ WARNING ⚠

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

    🥭 Entropy Markets: https://entropy.trade
    📄 Documentation: https://docs.entropy.trade/
    💬 Discord: https://discord.gg/67jySBhxrg
    🐦 Twitter: https://twitter.com/entropymarkets
    🚧 Github: https://github.com/blockworks-foundation
    📧 Email: mailto:hello@blockworks.foundation
"""


# # _build_data_path
#
# The code needs access to some data files, such as the ids.json file that's used in multiple Entropy projects. In
# this project all these data files are kept in the /data directory, relative to the project root.
#
# Some situations can make it tricky accessing files in that known location though. (In particular, Nuitka
# compilation to a standalone executable seems to make accessing internal paths with '..' in them impossible.)
#
# This function provides a consistent way to determine the correct data path for use throughout `entropy-explorer`.
#
def _build_data_path() -> str:
    possibilities: typing.Sequence[str] = [
        "../data",
        "data",
        ".",
        "../../data",
        "../../../data",
    ]
    attempts: typing.List[str] = []
    file_root: str = os.path.dirname(__file__)
    for possibility in possibilities:
        data_path: str = os.path.normpath(os.path.join(file_root, possibility))
        attempts += [data_path]
        try:
            attempted_ids_path: str = os.path.normpath(
                os.path.join(data_path, "ids.json")
            )
            with open(attempted_ids_path) as ids_file:
                json.load(ids_file)
                return data_path
        except:
            pass

    raise Exception(
        f"Could not determine data path - ids.json not found in: {attempts}"
    )


# # DATA_PATH
#
# This is the path to the data directory that contains (among other things) the ids.json.
#
DATA_PATH: str = _build_data_path()


# ## EntropyConstants
#
# Load all Entropy Market's constants from its own `ids.json` file (retrieved from [GitHub](https://raw.githubusercontent.com/blockworks-foundation/entropy-client-ts/main/src/ids.json).
#
with open(os.path.join(DATA_PATH, "ids.json")) as json_file:
    EntropyConstants = json.load(json_file)


# # 🥭 PackageVersion class
#
# Runtime details of the current version of entropy-explorer.
#
@dataclass
class PackageVersion:
    version: str
    last_commit: str

    def __str__(self) -> str:
        return f"« PackageVersion {self.version} - '{self.last_commit}' »"

    def __repr__(self) -> str:
        return f"{self}"


def version() -> PackageVersion:
    package_version: str = "Unknown"
    # The exception is deliberately trapped and ignored - we just want to return "Unknown" in that situation.
    try:  # nosemgrep
        package_version = importlib.metadata.version("entropy-explorer")
    except Exception:
        pass

    version_filename: str = os.path.join(DATA_PATH, ".version")
    last_commit = f"Unknown (no version file found at '{version_filename}')."
    if os.path.isfile(version_filename):
        with open(version_filename) as version_file:
            last_commit = version_file.read().strip()

    return PackageVersion(version=package_version, last_commit=last_commit)



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import enum
import entropy
import typing

from solana.publickey import PublicKey

from ..constants import SYSTEM_PROGRAM_ADDRESS
from ..modelstate import ModelState
from .modelstatebuilder import (
    ModelStateBuilder,
    WebsocketModelStateBuilder,
    SerumPollingModelStateBuilder,
    SpotPollingModelStateBuilder,
    PerpPollingModelStateBuilder,
)


class ModelUpdateMode(enum.Enum):
    # We use strings here so that argparse can work with these as parameters.
    WEBSOCKET = "WEBSOCKET"
    POLL = "POLL"

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 ModelStateBuilder class
#
# Base class for building a `ModelState` through polling or websockets.
#
def model_state_builder_factory(
    mode: ModelUpdateMode,
    context: entropy.Context,
    disposer: entropy.Disposable,
    websocket_manager: entropy.WebSocketSubscriptionManager,
    health_check: entropy.HealthCheck,
    wallet: entropy.Wallet,
    group: entropy.Group,
    account: entropy.Account,
    market: entropy.LoadedMarket,
    oracle: entropy.Oracle,
) -> ModelStateBuilder:
    if mode == ModelUpdateMode.WEBSOCKET:
        return _websocket_model_state_builder_factory(
            context,
            disposer,
            websocket_manager,
            health_check,
            wallet,
            group,
            account,
            market,
            oracle,
        )
    else:
        return _polling_model_state_builder_factory(
            context, wallet, group, account, market, oracle
        )


def _polling_model_state_builder_factory(
    context: entropy.Context,
    wallet: entropy.Wallet,
    group: entropy.Group,
    account: entropy.Account,
    market: entropy.Market,
    oracle: entropy.Oracle,
) -> ModelStateBuilder:
    if entropy.SerumMarket.isa(market):
        return _polling_serum_model_state_builder_factory(
            context, wallet, group, account, entropy.SerumMarket.ensure(market), oracle
        )
    elif entropy.SpotMarket.isa(market):
        return _polling_spot_model_state_builder_factory(
            group, account, entropy.SpotMarket.ensure(market), oracle
        )
    elif entropy.PerpMarket.isa(market):
        return _polling_perp_model_state_builder_factory(
            group, account, entropy.PerpMarket.ensure(market), oracle
        )
    else:
        raise Exception(
            f"Could not determine type of market {market.fully_qualified_symbol}: {market}"
        )


def _polling_serum_model_state_builder_factory(
    context: entropy.Context,
    wallet: entropy.Wallet,
    group: entropy.Group,
    account: entropy.Account,
    market: entropy.SerumMarket,
    oracle: entropy.Oracle,
) -> ModelStateBuilder:
    base_account = entropy.TokenAccount.fetch_largest_for_owner_and_token(
        context, wallet.address, market.base
    )
    if base_account is None:
        raise Exception(
            f"Could not find token account owned by {wallet.address} for base token {market.base}."
        )
    quote_account = entropy.TokenAccount.fetch_largest_for_owner_and_token(
        context, wallet.address, market.quote
    )
    if quote_account is None:
        raise Exception(
            f"Could not find token account owned by {wallet.address} for quote token {market.quote}."
        )
    all_open_orders = entropy.OpenOrders.load_for_market_and_owner(
        context,
        market.address,
        wallet.address,
        context.serum_program_address,
        market.base,
        market.quote,
    )
    if len(all_open_orders) == 0:
        raise Exception(
            f"Could not find serum openorders account owned by {wallet.address} for market {market.fully_qualified_symbol}."
        )
    return SerumPollingModelStateBuilder(
        all_open_orders[0].address,
        market,
        oracle,
        group.address,
        group.cache,
        account.address,
        all_open_orders[0].address,
        base_account,
        quote_account,
    )


def _polling_spot_model_state_builder_factory(
    group: entropy.Group,
    account: entropy.Account,
    market: entropy.SpotMarket,
    oracle: entropy.Oracle,
) -> ModelStateBuilder:
    market_index: int = group.slot_by_spot_market_address(market.address).index
    open_orders_address: typing.Optional[PublicKey] = account.spot_open_orders_by_index[
        market_index
    ]
    all_open_orders_addresses: typing.Sequence[PublicKey] = account.spot_open_orders
    if open_orders_address is None:
        raise Exception(
            f"Could not find spot openorders in account {account.address} for market {market.fully_qualified_symbol}."
        )
    return SpotPollingModelStateBuilder(
        open_orders_address,
        market,
        oracle,
        group.address,
        group.cache,
        account.address,
        open_orders_address,
        all_open_orders_addresses,
    )


def _polling_perp_model_state_builder_factory(
    group: entropy.Group,
    account: entropy.Account,
    market: entropy.PerpMarket,
    oracle: entropy.Oracle,
) -> ModelStateBuilder:
    all_open_orders_addresses: typing.Sequence[PublicKey] = account.spot_open_orders
    return PerpPollingModelStateBuilder(
        account.address,
        market,
        oracle,
        group.address,
        group.cache,
        all_open_orders_addresses,
    )


def __load_all_openorders_watchers(
    context: entropy.Context,
    wallet: entropy.Wallet,
    account: entropy.Account,
    group: entropy.Group,
    websocket_manager: entropy.WebSocketSubscriptionManager,
    health_check: entropy.HealthCheck,
) -> typing.Sequence[entropy.Watcher[entropy.OpenOrders]]:
    all_open_orders_watchers: typing.List[entropy.Watcher[entropy.OpenOrders]] = []
    for basket_token in account.base_slots:
        if basket_token.spot_open_orders is not None:
            spot_market_symbol: str = f"spot:{basket_token.base_instrument.symbol}/{account.shared_quote_token.symbol}"
            spot_market = entropy.SpotMarket.ensure(
                entropy.market(context, spot_market_symbol)
            )
            oo_watcher = entropy.build_spot_open_orders_watcher(
                context,
                websocket_manager,
                health_check,
                wallet,
                account,
                group,
                spot_market,
            )
            all_open_orders_watchers += [oo_watcher]

    return all_open_orders_watchers


def _websocket_model_state_builder_factory(
    context: entropy.Context,
    disposer: entropy.Disposable,
    websocket_manager: entropy.WebSocketSubscriptionManager,
    health_check: entropy.HealthCheck,
    wallet: entropy.Wallet,
    group: entropy.Group,
    account: entropy.Account,
    market: entropy.LoadedMarket,
    oracle: entropy.Oracle,
) -> ModelStateBuilder:
    group_watcher = entropy.build_group_watcher(
        context, websocket_manager, health_check, group
    )
    cache = entropy.Cache.load(context, group.cache)
    cache_watcher = entropy.build_cache_watcher(
        context, websocket_manager, health_check, cache, group
    )
    account_subscription, latest_account_observer = entropy.build_account_watcher(
        context, websocket_manager, health_check, account, group_watcher, cache_watcher
    )

    initial_price = oracle.fetch_price(context)
    price_feed = oracle.to_streaming_observable(context)
    latest_price_observer = entropy.LatestItemObserverSubscriber(initial_price)
    price_disposable = price_feed.subscribe(latest_price_observer)
    disposer.add_disposable(price_disposable)
    health_check.add("price_subscription", price_feed)

    if entropy.SerumMarket.isa(market):
        serum_market = entropy.SerumMarket.ensure(market)
        order_owner: PublicKey = (
            serum_market.find_openorders_address_for_owner(context, wallet.address)
            or SYSTEM_PROGRAM_ADDRESS
        )
        price_watcher: entropy.Watcher[entropy.Price] = entropy.build_price_watcher(
            context, websocket_manager, health_check, disposer, "market", serum_market
        )
        inventory_watcher: entropy.Watcher[
            entropy.Inventory
        ] = entropy.build_serum_inventory_watcher(
            context,
            websocket_manager,
            health_check,
            disposer,
            wallet,
            serum_market,
            price_watcher,
        )
        latest_open_orders_observer: entropy.Watcher[
            entropy.PlacedOrdersContainer
        ] = entropy.build_serum_open_orders_watcher(
            context, websocket_manager, health_check, serum_market, wallet
        )
        latest_orderbook_watcher: entropy.Watcher[
            entropy.OrderBook
        ] = entropy.build_orderbook_watcher(
            context, websocket_manager, health_check, serum_market
        )
        latest_event_queue_watcher: entropy.Watcher[
            entropy.EventQueue
        ] = entropy.build_serum_event_queue_watcher(
            context, websocket_manager, health_check, serum_market
        )
    elif entropy.SpotMarket.isa(market):
        spot_market = entropy.SpotMarket.ensure(market)
        market_index: int = group.slot_by_spot_market_address(spot_market.address).index
        order_owner = (
            account.spot_open_orders_by_index[market_index] or SYSTEM_PROGRAM_ADDRESS
        )

        all_open_orders_watchers = __load_all_openorders_watchers(
            context, wallet, account, group, websocket_manager, health_check
        )
        latest_open_orders_observer = list(
            [
                oo_watcher
                for oo_watcher in all_open_orders_watchers
                if (
                    spot_market.base == spot_market.base
                    and spot_market.quote == spot_market.quote
                )
            ]
        )[0]

        inventory_watcher = entropy.InventoryAccountWatcher(
            spot_market,
            latest_account_observer,
            group_watcher,
            all_open_orders_watchers,
            cache_watcher,
        )
        latest_orderbook_watcher = entropy.build_orderbook_watcher(
            context, websocket_manager, health_check, spot_market
        )
        latest_event_queue_watcher = entropy.build_spot_event_queue_watcher(
            context, websocket_manager, health_check, spot_market
        )
    elif entropy.PerpMarket.isa(market):
        perp_market = entropy.PerpMarket.ensure(market)
        order_owner = account.address

        all_open_orders_watchers = __load_all_openorders_watchers(
            context, wallet, account, group, websocket_manager, health_check
        )

        inventory_watcher = entropy.InventoryAccountWatcher(
            perp_market,
            latest_account_observer,
            group_watcher,
            all_open_orders_watchers,
            cache_watcher,
        )

        latest_open_orders_observer = entropy.build_perp_open_orders_watcher(
            context,
            websocket_manager,
            health_check,
            perp_market,
            account,
            group,
            account_subscription,
        )
        latest_orderbook_watcher = entropy.build_orderbook_watcher(
            context, websocket_manager, health_check, perp_market
        )
        latest_event_queue_watcher = entropy.build_perp_event_queue_watcher(
            context, websocket_manager, health_check, perp_market
        )
    else:
        raise Exception(
            f"Could not determine type of market {market.fully_qualified_symbol} - {market}"
        )

    model_state = ModelState(
        order_owner,
        market,
        group_watcher,
        latest_account_observer,
        latest_price_observer,
        latest_open_orders_observer,
        inventory_watcher,
        latest_orderbook_watcher,
        latest_event_queue_watcher,
    )
    return WebsocketModelStateBuilder(model_state)



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import abc
import enum
import logging
import rx
import typing

from datetime import datetime
from decimal import Decimal

from .context import Context
from .loadedmarket import LoadedMarket


# # 🥭 Oracles
#
# This file deals with fetching prices from exchanges and oracles.
#

# # 🥭 SupportedOracleFeature enum
#
# Some oracles provide a mid price. Some provide bid and ask but no mid price. Some provide a confidence score.
# Some done. This enum allows an oracle to say which features it supports.
#
class SupportedOracleFeature(enum.Flag):
    MID_PRICE = enum.auto()
    TOP_BID_AND_OFFER = enum.auto()
    CONFIDENCE = enum.auto()

    def has_feature(self, feature: "SupportedOracleFeature") -> bool:
        return (self & feature) == 0  # type: ignore[comparison-overlap]


# # 🥭 OracleSource class
#
# This class describes an oracle and can be used to tell apart `Prices` from different `Oracle`s
# apart.
#
class OracleSource:
    def __init__(
        self,
        provider_name: str,
        source_name: str,
        supports: SupportedOracleFeature,
        market: LoadedMarket,
    ) -> None:
        self.provider_name = provider_name
        self.source_name = source_name
        self.supports: SupportedOracleFeature = supports
        self.market = market

    def __str__(self) -> str:
        return f"« OracleSource '{self.source_name}' from '{self.provider_name}' for market '{self.market.fully_qualified_symbol}' [{self.supports}] »"

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 Price class
#
# This class contains all relevant info for a price.
#
class Price:
    def __init__(
        self,
        source: OracleSource,
        timestamp: datetime,
        market: LoadedMarket,
        top_bid: Decimal,
        mid_price: Decimal,
        top_ask: Decimal,
        confidence: Decimal,
    ) -> None:
        self.source: OracleSource = source
        self.timestamp: datetime = timestamp
        self.market: LoadedMarket = market
        self.top_bid: Decimal = top_bid
        self.mid_price: Decimal = mid_price
        self.top_ask: Decimal = top_ask
        self.confidence: Decimal = confidence

    @property
    def spread(self) -> Decimal:
        return (self.top_ask - self.top_bid) / 2

    def __str__(self) -> str:
        confidence = ""
        if self.source.supports & SupportedOracleFeature.CONFIDENCE:
            confidence = f" +/- {self.confidence:,.8f}"
        return f"« Price [{self.source.provider_name}] {self.market.fully_qualified_symbol} at {self.timestamp}: {self.mid_price:,.8f}{confidence} »"

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 Oracle class
#
# Derived versions of this class can fetch prices for a specific market.
#
class Oracle(metaclass=abc.ABCMeta):
    def __init__(self, name: str, market: LoadedMarket) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.name = name
        self.market = market

    @property
    def symbol(self) -> str:
        return self.market.symbol

    @abc.abstractmethod
    def fetch_price(self, context: Context) -> Price:
        raise NotImplementedError(
            "Oracle.fetch_price() is not implemented on the base type."
        )

    @abc.abstractmethod
    def to_streaming_observable(
        self, context: Context
    ) -> rx.core.typing.Observable[Price]:
        raise NotImplementedError(
            "Oracle.fetch_price() is not implemented on the base type."
        )

    def __str__(self) -> str:
        return f"« Oracle {self.name} [{self.market.fully_qualified_symbol}] »"

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 OracleProvider class
#
# Derived versions of this class allow creation of oracles for markets.
#
class OracleProvider(metaclass=abc.ABCMeta):
    def __init__(self, name: str) -> None:
        self.name = name

    @abc.abstractmethod
    def oracle_for_market(
        self, context: Context, market: LoadedMarket
    ) -> typing.Optional[Oracle]:
        raise NotImplementedError(
            "OracleProvider.create_oracle_for_market() is not implemented on the base type."
        )

    @abc.abstractmethod
    def all_available_symbols(self, context: Context) -> typing.Sequence[str]:
        raise NotImplementedError(
            "OracleProvider.all_available_symbols() is not implemented on the base type."
        )

    def __str__(self) -> str:
        return f"« OracleProvider {self.name} »"

    def __repr__(self) -> str:
        return f"{self}"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import logging
import typing

from .version import Version


# # 🥭 Metadata class
#
class Metadata:
    def __init__(
        self, data_type: typing.Any, version: Version, is_initialized: bool
    ) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.data_type: typing.Any = data_type
        self.version: Version = version
        self.is_initialized: bool = is_initialized

    def from_layout(layout: typing.Any) -> "Metadata":
        version = Version(layout.version + 1)
        is_initialized = bool(layout.is_initialized)
        return Metadata(layout.data_type, version, is_initialized)

    def __str__(self) -> str:
        init = "Initialized" if self.is_initialized else "Not Initialized"
        return f"« Metadata {self.version} - {self.data_type}: {init} »"

    def __repr__(self) -> str:
        return f"{self}"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import abc
import logging
import entropy

from datetime import datetime

from ..observables import EventSource


# # 🥭 Hedger class
#
# A base hedger class to allow hedging across markets.
#
class Hedger(metaclass=abc.ABCMeta):
    def __init__(self) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.pulse_complete: EventSource[datetime] = EventSource[datetime]()
        self.pulse_error: EventSource[Exception] = EventSource[Exception]()

    @abc.abstractmethod
    def pulse(self, context: entropy.Context, model_state: entropy.ModelState) -> None:
        raise NotImplementedError("Hedger.pulse() is not implemented on the base type.")

    def __str__(self) -> str:
        return "« Hedger »"

    def __repr__(self) -> str:
        return f"{self}"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import logging
import typing

from decimal import Decimal
from solana.publickey import PublicKey

from .account import Account
from .group import Group
from .inventory import Inventory
from .markets import Market
from .oracle import Price
from .orders import Order, OrderBook
from .placedorder import PlacedOrdersContainer
from .watcher import Watcher


# # 🥭 EventQueue protocol
#
# The `EventQueue` protocol just says the object has a property `accounts_to_crank` that returns a `Sequence` of `PublicKey`s.
#
# This is to share the interface between a `SerumEventQueue` and a `PerpEventQueue` for cranking, when the
# underlying objects are really quite different.
#
class EventQueue(typing.Protocol):
    @property
    def accounts_to_crank(self) -> typing.Sequence[PublicKey]:
        raise NotImplementedError(
            "EventQueue.accounts_to_crank is not implemented on the Protocol."
        )


class NullEventQueue:
    @property
    def accounts_to_crank(self) -> typing.Sequence[PublicKey]:
        return []


# # 🥭 ModelState class
#
# Provides simple access to the latest state of market and account data.
#
class ModelState:
    def __init__(
        self,
        order_owner: PublicKey,
        market: Market,
        group_watcher: Watcher[Group],
        account_watcher: Watcher[Account],
        price_watcher: Watcher[Price],
        placed_orders_container_watcher: Watcher[PlacedOrdersContainer],
        inventory_watcher: Watcher[Inventory],
        orderbook: Watcher[OrderBook],
        event_queue: Watcher[EventQueue],
    ) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.order_owner: PublicKey = order_owner
        self.market: Market = market
        self.group_watcher: Watcher[Group] = group_watcher
        self.account_watcher: Watcher[Account] = account_watcher
        self.price_watcher: Watcher[Price] = price_watcher
        self.placed_orders_container_watcher: Watcher[
            PlacedOrdersContainer
        ] = placed_orders_container_watcher
        self.inventory_watcher: Watcher[Inventory] = inventory_watcher
        self.orderbook_watcher: Watcher[OrderBook] = orderbook
        self.event_queue_watcher: Watcher[EventQueue] = event_queue

        self.not_quoting: bool = False
        self.state: typing.Dict[str, typing.Any] = {}

    @property
    def group(self) -> Group:
        return self.group_watcher.latest

    @property
    def account(self) -> Account:
        return self.account_watcher.latest

    @property
    def price(self) -> Price:
        return self.price_watcher.latest

    @property
    def placed_orders_container(self) -> PlacedOrdersContainer:
        return self.placed_orders_container_watcher.latest

    @property
    def inventory(self) -> Inventory:
        return self.inventory_watcher.latest

    @property
    def orderbook(self) -> OrderBook:
        return self.orderbook_watcher.latest

    @property
    def bids(self) -> typing.Sequence[Order]:
        return self.orderbook.bids

    @property
    def asks(self) -> typing.Sequence[Order]:
        return self.orderbook.asks

    # The top bid is the highest price someone is willing to pay to BUY
    @property
    def top_bid(self) -> typing.Optional[Order]:
        return self.orderbook.top_bid

    # The top ask is the lowest price someone is willing to pay to SELL
    @property
    def top_ask(self) -> typing.Optional[Order]:
        return self.orderbook.top_ask

    @property
    def spread(self) -> Decimal:
        return self.orderbook.spread

    @property
    def accounts_to_crank(self) -> typing.Sequence[PublicKey]:
        return self.event_queue_watcher.latest.accounts_to_crank

    def current_orders(self) -> typing.Sequence[Order]:
        return self.orderbook.all_orders_for_owner(self.order_owner, cutoff=None)

    def __str__(self) -> str:
        return f"""« ModelState for market '{self.market.fully_qualified_symbol}'
    Group: {self.group_watcher.latest.address}
    Account: {self.account_watcher.latest.address}
    Price: {self.price_watcher.latest}
    Inventory: {self.inventory_watcher.latest}
    Existing Order Count: {len(self.placed_orders_container_watcher.latest.placed_orders)}
    Bid Count: {len(self.bids)}
    Ask Count: {len(self.asks)}
»"""

    def __repr__(self) -> str:
        return f"{self}"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import argparse
import entropy
import typing

from decimal import Decimal

from .pairwiseelement import PairwiseElement
from ...modelstate import ModelState


# # 🥭 BiasQuoteElement class
#
# Modifies an `Order`s price based. Uses `bias` to shift the price down to sell more (using
# a `biasquote-bias` of less than 1) or buy more (using a `biasquote-bias` of greater than 1).
#
# Can take multiple bias factors to work with pair-wise orders.
#
class BiasQuoteElement(PairwiseElement):
    def __init__(self, factors: typing.Sequence[Decimal]) -> None:
        super().__init__()
        self.bias_factors: typing.Sequence[Decimal] = factors

    @staticmethod
    def add_command_line_parameters(parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--biasquote-factor",
            type=Decimal,
            action="append",
            help="bias factor to apply to quotes. Prices will be multiplied by this factor, so a number less than 1 will reduce prices and a number greater than 1 will increase prices. For example, use 1.001 to increase prices by 10 bips. Can be specified multiple times to apply to different levels.",
        )

    @staticmethod
    def from_command_line_parameters(args: argparse.Namespace) -> "BiasQuoteElement":
        bias_factors: typing.Sequence[Decimal] = args.biasquote_factor
        return BiasQuoteElement(bias_factors or [Decimal(1)])

    def process_order_pair(
        self,
        context: entropy.Context,
        model_state: ModelState,
        index: int,
        buy: typing.Optional[entropy.Order],
        sell: typing.Optional[entropy.Order],
    ) -> typing.Tuple[typing.Optional[entropy.Order], typing.Optional[entropy.Order]]:
        # If no bias is explicitly specified for this element, just use the last specified bias.
        bias_factor: Decimal = (
            self.bias_factors[index]
            if index < len(self.bias_factors)
            else self.bias_factors[-1]
        )
        bias_description = "BUY more" if bias_factor > 1 else "SELL more"

        if bias_factor == 1:
            # Bias factor of 1 results in no changes to orders.
            return buy, sell

        new_buy: typing.Optional[entropy.Order] = None
        new_sell: typing.Optional[entropy.Order] = None
        if buy is not None:
            new_buy_price: Decimal = buy.price * bias_factor
            new_buy = buy.with_update(price=new_buy_price)
            self._logger.debug(
                f"""Order change - bias factor of {bias_factor} shifted price to {bias_description}:
    Old: {buy}
    New: {new_buy}"""
            )

        if sell is not None:
            new_sell_price: Decimal = sell.price * bias_factor
            new_sell = sell.with_update(price=new_sell_price)
            self._logger.debug(
                f"""Order change - bias factor of {bias_factor} shifted price to {bias_description}:
    Old: {sell}
    New: {new_sell}"""
            )

        return new_buy, new_sell

    def __str__(self) -> str:
        return f"« BiasQuoteElement - bias factors: {self.bias_factors} »"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import logging
import typing

from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from solana.publickey import PublicKey

from .accountinfo import AccountInfo
from .addressableaccount import AddressableAccount
from .cache import RootBankCache, Cache
from .context import Context
from .instrumentlookup import InstrumentLookup
from .layouts import layouts
from .metadata import Metadata
from .observables import Disposable
from .tokens import Instrument, Token
from .version import Version
from .websocketsubscription import (
    WebSocketAccountSubscription,
    WebSocketSubscriptionManager,
)


# # 🥭 InterestRates class
#
# A simple way to package borrow and deposit rates together in a single object.
#
@dataclass
class InterestRates:
    deposit: Decimal
    borrow: Decimal

    def __str__(self) -> str:
        return (
            f"« InterestRates Deposit: {self.deposit:,.2%} Borrow: {self.borrow:,.2%} »"
        )

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 BankBalances class
#
# A simple way to package borrow and deposit balances together in a single object.
#
@dataclass
class BankBalances:
    deposits: Decimal
    borrows: Decimal

    def __str__(self) -> str:
        return f"« BankBalances Deposits: {self.deposits:,.8f} Borrows: {self.borrows:,.8f} »"

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 NodeBank class
#
# `NodeBank` stores details of deposits/borrows and vault.
#
class NodeBank(AddressableAccount):
    def __init__(
        self,
        account_info: AccountInfo,
        version: Version,
        meta_data: Metadata,
        vault: PublicKey,
        balances: BankBalances,
    ) -> None:
        super().__init__(account_info)
        self.version: Version = version
        self.meta_data: Metadata = meta_data
        self.vault: PublicKey = vault
        self.balances: BankBalances = balances

    @staticmethod
    def from_layout(
        layout: typing.Any, account_info: AccountInfo, version: Version
    ) -> "NodeBank":
        meta_data: Metadata = layout.meta_data
        deposits: Decimal = layout.deposits
        borrows: Decimal = layout.borrows
        balances: BankBalances = BankBalances(deposits=deposits, borrows=borrows)
        vault: PublicKey = layout.vault

        return NodeBank(account_info, version, meta_data, vault, balances)

    @staticmethod
    def parse(account_info: AccountInfo) -> "NodeBank":
        data = account_info.data
        if len(data) != layouts.NODE_BANK.sizeof():
            raise Exception(
                f"NodeBank data length ({len(data)}) does not match expected size ({layouts.NODE_BANK.sizeof()})"
            )

        layout = layouts.NODE_BANK.parse(data)
        return NodeBank.from_layout(layout, account_info, Version.V1)

    @staticmethod
    def load(context: Context, address: PublicKey) -> "NodeBank":
        account_info = AccountInfo.load(context, address)
        if account_info is None:
            raise Exception(f"NodeBank account not found at address '{address}'")
        return NodeBank.parse(account_info)

    def subscribe(
        self,
        context: Context,
        websocketmanager: WebSocketSubscriptionManager,
        callback: typing.Callable[["NodeBank"], None],
    ) -> Disposable:
        subscription = WebSocketAccountSubscription(
            context, self.address, NodeBank.parse
        )
        websocketmanager.add(subscription)
        subscription.publisher.subscribe(on_next=callback)  # type: ignore[call-arg]

        return subscription

    def __str__(self) -> str:
        return f"""« NodeBank [{self.version}] {self.address}
    {self.meta_data}
    Balances: {self.balances}
    Vault: {self.vault}
»"""

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 RootBank class
#
# `RootBank` stores details of how to reach `NodeBank`.
#
class RootBank(AddressableAccount):
    def __init__(
        self,
        account_info: AccountInfo,
        version: Version,
        meta_data: Metadata,
        optimal_util: Decimal,
        optimal_rate: Decimal,
        max_rate: Decimal,
        node_banks: typing.Sequence[PublicKey],
        deposit_index: Decimal,
        borrow_index: Decimal,
        last_updated: datetime,
    ) -> None:
        super().__init__(account_info)
        self.version: Version = version

        self.meta_data: Metadata = meta_data

        self.optimal_util: Decimal = optimal_util
        self.optimal_rate: Decimal = optimal_rate
        self.max_rate: Decimal = max_rate

        self.node_banks: typing.Sequence[PublicKey] = node_banks
        self.deposit_index: Decimal = deposit_index
        self.borrow_index: Decimal = borrow_index
        self.last_updated: datetime = last_updated

        self.loaded_node_banks: typing.Optional[typing.Sequence[NodeBank]] = None

    def ensure_node_banks(self, context: Context) -> typing.Sequence[NodeBank]:
        if self.loaded_node_banks is None:
            node_bank_account_infos = AccountInfo.load_multiple(
                context, self.node_banks
            )
            self.loaded_node_banks = list(map(NodeBank.parse, node_bank_account_infos))
        return self.loaded_node_banks

    def pick_node_bank(self, context: Context) -> NodeBank:
        return self.ensure_node_banks(context)[0]

    def fetch_balances(self, context: Context) -> BankBalances:
        node_banks: typing.Sequence[NodeBank] = self.ensure_node_banks(context)

        deposits_in_node_banks: Decimal = Decimal(0)
        borrows_in_node_banks: Decimal = Decimal(0)
        for node_bank in node_banks:
            deposits_in_node_banks += node_bank.balances.deposits
            borrows_in_node_banks += node_bank.balances.borrows

        total_deposits: Decimal = deposits_in_node_banks * self.deposit_index
        total_borrows: Decimal = borrows_in_node_banks * self.borrow_index

        return BankBalances(deposits=total_deposits, borrows=total_borrows)

    @staticmethod
    def from_layout(
        layout: typing.Any, account_info: AccountInfo, version: Version
    ) -> "RootBank":
        meta_data: Metadata = Metadata.from_layout(layout.meta_data)

        optimal_util: Decimal = layout.optimal_util
        optimal_rate: Decimal = layout.optimal_rate
        max_rate: Decimal = layout.max_rate

        num_node_banks: Decimal = layout.num_node_banks
        node_banks: typing.Sequence[PublicKey] = layout.node_banks[
            0 : int(num_node_banks)
        ]
        deposit_index: Decimal = layout.deposit_index
        borrow_index: Decimal = layout.borrow_index
        last_updated: datetime = layout.last_updated

        return RootBank(
            account_info,
            version,
            meta_data,
            optimal_util,
            optimal_rate,
            max_rate,
            node_banks,
            deposit_index,
            borrow_index,
            last_updated,
        )

    @staticmethod
    def parse(account_info: AccountInfo) -> "RootBank":
        data = account_info.data
        if len(data) != layouts.ROOT_BANK.sizeof():
            raise Exception(
                f"RootBank data length ({len(data)}) does not match expected size ({layouts.ROOT_BANK.sizeof()})"
            )

        layout = layouts.ROOT_BANK.parse(data)
        return RootBank.from_layout(layout, account_info, Version.V1)

    @staticmethod
    def load(context: Context, address: PublicKey) -> "RootBank":
        account_info = AccountInfo.load(context, address)
        if account_info is None:
            raise Exception(f"RootBank account not found at address '{address}'")
        return RootBank.parse(account_info)

    @staticmethod
    def load_multiple(
        context: Context, addresses: typing.Sequence[PublicKey]
    ) -> typing.Sequence["RootBank"]:
        account_infos = AccountInfo.load_multiple(context, addresses)
        root_banks = []
        for account_info in account_infos:
            root_bank = RootBank.parse(account_info)
            root_banks += [root_bank]

        return root_banks

    @staticmethod
    def find_by_address(
        values: typing.Sequence["RootBank"], address: PublicKey
    ) -> "RootBank":
        found = [value for value in values if value.address == address]
        if len(found) == 0:
            raise Exception(f"RootBank '{address}' not found in root banks: {values}")

        if len(found) > 1:
            raise Exception(
                f"RootBank '{address}' matched multiple root banks in: {values}"
            )

        return found[0]

    def subscribe(
        self,
        context: Context,
        websocketmanager: WebSocketSubscriptionManager,
        callback: typing.Callable[["RootBank"], None],
    ) -> Disposable:
        subscription = WebSocketAccountSubscription(
            context, self.address, RootBank.parse
        )
        websocketmanager.add(subscription)
        subscription.publisher.subscribe(on_next=callback)  # type: ignore[call-arg]

        return subscription

    def __str__(self) -> str:
        return f"""« RootBank [{self.version}] {self.address}
    {self.meta_data}
    Optimal Util: {self.optimal_util:,.4f}
    Optimal Rate: {self.optimal_rate:,.4f}
    Max Rate: {self.max_rate}
    Node Banks:
        {self.node_banks}
    Deposit Index: {self.deposit_index}
    Borrow Index: {self.borrow_index}
    Last Updated: {self.last_updated}
»"""

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 TokenBank class
#
# `TokenBank` defines additional information for a `Token`.
#
class TokenBank:
    def __init__(self, token: Token, root_bank_address: PublicKey) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.token: Token = token
        self.root_bank_address: PublicKey = root_bank_address

        self.loaded_root_bank: typing.Optional[RootBank] = None

    @staticmethod
    def from_layout_or_none(
        layout: typing.Any, instrument_lookup: InstrumentLookup
    ) -> typing.Optional["TokenBank"]:
        if layout.mint is None:
            return None

        instrument: typing.Optional[Instrument] = instrument_lookup.find_by_mint(
            layout.mint
        )
        if instrument is None:
            raise Exception(f"Token with mint {layout.mint} could not be found.")
        token: Token = Token.ensure(instrument)
        root_bank_address: PublicKey = layout.root_bank
        decimals: Decimal = layout.decimals

        if decimals != token.decimals:
            raise Exception(
                f"Conflict between number of decimals in token static data {token.decimals} and group {decimals} for token {token.symbol}."
            )

        return TokenBank(token, root_bank_address)

    @staticmethod
    def find_by_symbol(
        values: typing.Sequence[typing.Optional["TokenBank"]], symbol: str
    ) -> "TokenBank":
        found = [
            value
            for value in values
            if value is not None
            and value.token is not None
            and value.token.symbol_matches(symbol)
        ]
        if len(found) == 0:
            raise Exception(f"Token '{symbol}' not found in token infos: {values}")

        if len(found) > 1:
            raise Exception(
                f"Token '{symbol}' matched multiple tokens in infos: {values}"
            )

        return found[0]

    def root_bank_cache_from_cache(
        self, cache: Cache, index: int
    ) -> typing.Optional[RootBankCache]:
        return cache.root_bank_cache[index]

    def ensure_root_bank(self, context: Context) -> RootBank:
        if self.loaded_root_bank is None:
            self.loaded_root_bank = RootBank.load(context, self.root_bank_address)
        return self.loaded_root_bank

    def pick_node_bank(self, context: Context) -> NodeBank:
        root_bank: RootBank = self.ensure_root_bank(context)
        return root_bank.pick_node_bank(context)

    def fetch_interest_rates(self, context: Context) -> InterestRates:
        root_bank: RootBank = self.ensure_root_bank(context)
        balances: BankBalances = root_bank.fetch_balances(context)

        borrow_rate: Decimal = Decimal(0)
        deposit_rate: Decimal = Decimal(0)
        utilization: Decimal
        if balances.deposits != 0 and balances.borrows != 0:
            if balances.deposits <= balances.borrows:
                borrow_rate = root_bank.max_rate
            else:
                utilization = balances.borrows / balances.deposits
                slope: Decimal
                if utilization < root_bank.optimal_util:
                    slope = root_bank.optimal_rate / root_bank.optimal_util
                    borrow_rate = slope * utilization
                else:
                    extra_utilization = utilization - root_bank.optimal_util
                    slope = (root_bank.max_rate - root_bank.optimal_rate) / (
                        1 - root_bank.optimal_util
                    )
                    borrow_rate = root_bank.optimal_rate + (slope * extra_utilization)

            if balances.deposits == 0:
                deposit_rate = root_bank.max_rate
            else:
                utilization = balances.borrows / balances.deposits
                deposit_rate = utilization * borrow_rate

        return InterestRates(deposit=deposit_rate, borrow=borrow_rate)

    def __str__(self) -> str:
        return f"""« TokenBank {self.token}
    Root Bank Address: {self.root_bank_address}
»"""

    def __repr__(self) -> str:
        return f"{self}"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import argparse
import entropy
import typing

from .element import Element
from ...modelstate import ModelState


# # 🥭 QuoteSingleSideElement class
#
# Only allows orders from one side of the book to progress to the next element of the chain.
#
class QuoteSingleSideElement(Element):
    def __init__(self, side: entropy.Side) -> None:
        super().__init__()
        self.allowed: entropy.Side = side

    @staticmethod
    def add_command_line_parameters(parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--quotesingleside-side",
            type=entropy.Side,
            help="the single side to quote on - if BUY, all SELLs will be removed from desired orders, if SELL, all BUYs will be removed.",
        )

    @staticmethod
    def from_command_line_parameters(
        args: argparse.Namespace,
    ) -> "QuoteSingleSideElement":
        side: entropy.Side = args.quotesingleside_side
        return QuoteSingleSideElement(side)

    def process(
        self,
        context: entropy.Context,
        model_state: ModelState,
        orders: typing.Sequence[entropy.Order],
    ) -> typing.Sequence[entropy.Order]:
        new_orders: typing.List[entropy.Order] = []
        for order in orders:
            if order.side == self.allowed:
                self._logger.debug(
                    f"""Allowing {order.side} order [allowed: {self.allowed}]:
    Allowed: {order}"""
                )
                new_orders += [order]
            else:
                self._logger.debug(
                    f"""Removing {order.side} order [allowed: {self.allowed}]:
    Removed: {order}"""
                )

        return new_orders

    def __str__(self) -> str:
        return "« QuoteSingleSideElement »"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import abc
import logging
import typing

from solana.publickey import PublicKey

from .markets import Market


# # 🥭 MarketLookup class
#
# This base class allows specialised ways of looking up markets by symbol or by address.
#


class MarketLookup(metaclass=abc.ABCMeta):
    def __init__(self) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)

    @abc.abstractmethod
    def find_by_symbol(self, symbol: str) -> typing.Optional[Market]:
        raise NotImplementedError(
            "MarketLookup.find_by_symbol() is not implemented on the base type."
        )

    @abc.abstractmethod
    def find_by_address(self, address: PublicKey) -> typing.Optional[Market]:
        raise NotImplementedError(
            "MarketLookup.find_by_address() is not implemented on the base type."
        )

    @abc.abstractmethod
    def all_markets(self) -> typing.Sequence[Market]:
        raise NotImplementedError(
            "MarketLookup.all_markets() is not implemented on the base type."
        )


# # 🥭 NullMarketLookup class
#
# This class is a simple stub `MarketLookup` that never returns a `Market`.
#


class NullMarketLookup(MarketLookup):
    def __init__(self) -> None:
        super().__init__()

    def find_by_symbol(self, symbol: str) -> typing.Optional[Market]:
        return None

    def find_by_address(self, address: PublicKey) -> typing.Optional[Market]:
        return None

    def all_markets(self) -> typing.Sequence[Market]:
        return []


# # 🥭 CompoundMarketLookup class
#
# This class allows multiple `MarketLookup` objects to be combined, returning the first valid lookup result
# found.
#


class CompoundMarketLookup(MarketLookup):
    def __init__(self, lookups: typing.Sequence[MarketLookup]) -> None:
        super().__init__()
        self.lookups: typing.Sequence[MarketLookup] = lookups

    def find_by_symbol(self, symbol: str) -> typing.Optional[Market]:
        for lookup in self.lookups:
            result = lookup.find_by_symbol(symbol)
            if result is not None:
                return result
        return None

    def find_by_address(self, address: PublicKey) -> typing.Optional[Market]:
        for lookup in self.lookups:
            result = lookup.find_by_address(address)
            if result is not None:
                return result
        return None

    def all_markets(self) -> typing.Sequence[Market]:
        return [
            market
            for sublist in map(lambda lookup: lookup.all_markets(), self.lookups)
            for market in sublist
        ]



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import json
import logging
import rx
import rx.subject
import threading
import time
import traceback
import typing
import websocket

from threading import Thread

from .datetimes import local_now


# # 🥭 ReconnectingWebsocket class
#
# The `ReconnectingWebsocket` class wraps a websocket with an automatic-reconnection
# mechanism. If an error disconnects the websocket, it will automatically reconnect. It
# will continue to automatically reconnect, until it is explicitly closed.
#
class ReconnectingWebsocket:
    def __init__(
        self, url: str, on_open_call: typing.Callable[[websocket.WebSocketApp], None]
    ) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.url = url
        self.on_open_call = on_open_call
        self.reconnect_required: bool = True
        self.pause_on_error: float = 0.0
        self.ping_interval: int = 0
        self.connecting: rx.subject.behaviorsubject.BehaviorSubject = (
            rx.subject.behaviorsubject.BehaviorSubject(local_now())
        )
        self.connected: rx.subject.behaviorsubject.BehaviorSubject = (
            rx.subject.behaviorsubject.BehaviorSubject(local_now())
        )
        self.disconnected: rx.subject.behaviorsubject.BehaviorSubject = (
            rx.subject.behaviorsubject.BehaviorSubject(local_now())
        )
        self.ping: rx.subject.behaviorsubject.BehaviorSubject = (
            rx.subject.behaviorsubject.BehaviorSubject(local_now())
        )
        self.pong: rx.subject.behaviorsubject.BehaviorSubject = (
            rx.subject.behaviorsubject.BehaviorSubject(local_now())
        )
        self.item: rx.subject.subject.Subject = rx.subject.subject.Subject()

        self.__open_event: threading.Event = threading.Event()

    def close(self) -> None:
        self._logger.debug(f"Closing WebSocket for {self.url}")
        self.reconnect_required = False
        self._ws.close()

    def force_reconnect(self) -> None:
        self._logger.debug(f"Forcing a reconnect on WebSocket for {self.url}")
        self._ws.close()

    def _on_open(self, ws: websocket.WebSocketApp) -> None:
        self._logger.debug(f"Opening WebSocket for {self.url}")
        self.__open_event.set()
        if self.on_open_call:
            self.on_open_call(ws)

    def _on_message(self, _: typing.Any, message: str) -> None:
        try:
            data = json.loads(message)
            self.item.on_next(data)
        except Exception:
            self._logger.error(f"Problem sending update: {traceback.format_exc()}")

    def _on_error(self, *args: typing.Any) -> None:
        self._logger.warning(f"WebSocket for {self.url} has error {args}")

    def _on_ping(self, *_: typing.Any) -> None:
        self.ping.on_next(local_now())

    def _on_pong(self, *_: typing.Any) -> None:
        self.pong.on_next(local_now())

    def open(self) -> None:
        thread = Thread(target=self._run)
        thread.start()

    def send(self, message: str) -> None:
        self._ws.send(message)

    def wait_until_open(self, timeout: float) -> bool:
        return self.__open_event.wait(timeout)

    def _run(self) -> None:
        while self.reconnect_required:
            try:
                self._logger.debug(f"WebSocket connecting to: {self.url}")
                self.connecting.on_next(local_now())
                self._ws = websocket.WebSocketApp(
                    self.url,
                    on_open=self._on_open,
                    on_message=self._on_message,
                    on_error=self._on_error,
                    on_ping=self._on_ping,
                    on_pong=self._on_pong,
                )
                self.connected.on_next(local_now())
                self._ws.run_forever(ping_interval=self.ping_interval)
                self.__open_event.clear()
                self.disconnected.on_next(local_now())
            except Exception:
                self._logger.warning(
                    f"Websocket received error: {traceback.format_exc()}"
                )
                if self.pause_on_error > 0:
                    self._logger.warning(
                        f"Pausing for {self.pause_on_error} seconds before trying to reconnect."
                    )
                    time.sleep(self.pause_on_error)



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://markets/) support is available at:
#   [Docs](https://docs.markets/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import re
import rx
import rx.operators
import typing

from decimal import Decimal
from solana.publickey import PublicKey

from ...accountinfo import AccountInfo
from ...context import Context
from ...datetimes import utc_now
from ...loadedmarket import LoadedMarket
from ...observables import observable_pipeline_error_reporter
from ...oracle import (
    Oracle,
    OracleProvider,
    OracleSource,
    Price,
    SupportedOracleFeature,
)

from .layouts import (
    MAGIC,
    MAPPING,
    PRICE,
    PRODUCT,
    PYTH_DEVNET_MAPPING_ROOT,
    PYTH_MAINNET_MAPPING_ROOT,
)


# # 🥭 Pyth
#
# This file contains code specific to the [Pyth Network](https://pyth.network/).
#
# ## Pyth's Confidence Interval
#
# It seems to akin to a spread, and always specified in the quote currency.
#
# Pyth Medium article:
# https://pythnetwork.medium.com/what-is-confidence-uncertainty-in-a-price-649583b598cf
#
# From Discord: https://discord.com/channels/826115122799837205/861627739975319602/861903479331880981
# 6/ What does Pyth's confidence value mean when provided with the price?
# → Confidence is a function of where one expects a trade to occur from the "true" price, i.e. what is the potential dispersion estimation?
#
# From Discord: https://discord.com/channels/826115122799837205/826115122799837208/858091212980092978
# [T]he confidence value here is how far from the aggregate price we believe the true price might be. It
# reflects a combination of the uncertainty of individual quoters and how well individual quoters agree
# with each other
#


# # 🥭 PythOracle class
#
# Implements the `Oracle` abstract base class specialised to the Pyth Network.
#


class PythOracle(Oracle):
    def __init__(
        self, context: Context, market: LoadedMarket, product_data: typing.Any
    ):
        name = f"Pyth Oracle for {market.fully_qualified_symbol}"
        super().__init__(name, market)
        self.context: Context = context
        self.market: LoadedMarket = market
        self.product_data: typing.Any = product_data
        self.address: PublicKey = product_data.address
        features: SupportedOracleFeature = (
            SupportedOracleFeature.MID_PRICE | SupportedOracleFeature.CONFIDENCE
        )
        self.source: OracleSource = OracleSource("Pyth", name, features, market)

    def fetch_price(self, _: Context) -> Price:
        price_account_info = AccountInfo.load(self.context, self.product_data.px_acc)
        if price_account_info is None:
            raise Exception(
                f"[{self.context.name}] Price account {self.product_data.px_acc} not found."
            )

        if len(price_account_info.data) != PRICE.sizeof():
            raise Exception(
                f"[{self.context.name}] Price account data has incorrect size. Expected: {PRICE.sizeof()}, got {len(price_account_info.data)}."
            )

        price_data = PRICE.parse(price_account_info.data)
        if price_data.magic != MAGIC:
            raise Exception(
                f"[{self.context.name}] Price account {price_account_info.address} is not a Pyth account."
            )

        factor = Decimal(10) ** price_data.expo
        price = price_data.agg.price * factor
        confidence = price_data.agg.conf * factor

        # Pyth has no notion of bids, asks, or spreads so just provide the single price.
        return Price(
            self.source,
            utc_now(),
            self.market,
            price,
            price,
            price,
            confidence,
        )

    def to_streaming_observable(
        self, context: Context
    ) -> rx.core.typing.Observable[Price]:
        prices = rx.interval(1).pipe(
            rx.operators.observe_on(context.create_thread_pool_scheduler()),
            rx.operators.start_with(-1),
            rx.operators.map(lambda _: self.fetch_price(context)),
            rx.operators.catch(observable_pipeline_error_reporter),
            rx.operators.retry(),
        )
        return typing.cast(rx.core.typing.Observable[Price], prices)


# # 🥭 PythOracleProvider class
#
# Implements the `OracleProvider` abstract base class specialised to the Pyth Network.
#
# In order to allow it to vary its cluster without affecting other programs, this takes a `Context` in its
# constructor and uses that to access the data. It ignores the context passed as a parameter to its methods.
# This allows the context-fudging to only happen on construction.


class PythOracleProvider(OracleProvider):
    def __init__(self, context: Context) -> None:
        self.address: PublicKey = (
            PYTH_MAINNET_MAPPING_ROOT
            if context.client.cluster_name == "mainnet"
            else PYTH_DEVNET_MAPPING_ROOT
        )
        self.__symbol_prefix = "Crypto."

        super().__init__(f"Pyth Oracle Factory [{self.address}]")
        self.context: Context = context

    def oracle_for_market(
        self, _: Context, market: LoadedMarket
    ) -> typing.Optional[Oracle]:
        pyth_symbol = self._market_symbol_to_pyth_symbol(market.symbol)
        products = self._fetch_all_pyth_products(self.context, self.address)
        for product in products:
            if product.attr["symbol"] == pyth_symbol:
                return PythOracle(self.context, market, product)
        return None

    def all_available_symbols(self, _: Context) -> typing.Sequence[str]:
        products = self._fetch_all_pyth_products(self.context, self.address)
        symbols: typing.List[str] = []
        for product in products:
            symbol = product.attr["symbol"]
            symbols += self._pyth_symbol_to_market_symbols(symbol)
        return symbols

    def _market_symbol_to_pyth_symbol(self, symbol: str) -> str:
        normalised = symbol.upper()
        prefixed = self.__symbol_prefix + normalised
        fixed_usdt = re.sub("USDT$", "USD", prefixed)
        fixed_usdc = re.sub("USDC$", "USD", fixed_usdt)
        fixed_perp = re.sub("\\-PERP$", "/USD", fixed_usdc)
        return fixed_perp

    def _pyth_symbol_to_market_symbols(self, symbol: str) -> typing.Sequence[str]:
        if symbol.endswith("USD"):
            return [f"{symbol}C", f"{symbol}T"]
        return [symbol]

    def _load_pyth_mapping(self, context: Context, address: PublicKey) -> typing.Any:
        account_info = AccountInfo.load(context, address)
        if account_info is None:
            raise Exception(
                f"[{context.name}] Pyth mapping account {address} not found."
            )

        if len(account_info.data) != MAPPING.sizeof():
            raise Exception(
                f"Mapping account data has incorrect size. Expected: {MAPPING.sizeof()}, got {len(account_info.data)}."
            )

        mapping: typing.Any = MAPPING.parse(account_info.data)
        mapping.address = account_info.address
        if mapping.magic != MAGIC:
            raise Exception(
                f"[{context.name}] Mapping account {account_info.address} is not a Pyth account."
            )

        return mapping

    def _fetch_all_pyth_products(
        self, context: Context, address: PublicKey
    ) -> typing.Sequence[typing.Any]:
        mapping = self._load_pyth_mapping(context, address)
        all_product_addresses = mapping.products[0 : int(mapping.num)]
        product_account_infos = AccountInfo.load_multiple(
            context, all_product_addresses
        )
        products: typing.List[typing.Any] = []
        for product_account_info in product_account_infos:
            product: typing.Any = PRODUCT.parse(product_account_info.data)
            product.address = product_account_info.address
            if product.magic != MAGIC:
                raise Exception(
                    f"[{context.name}] Product account {product_account_info.address} is not a Pyth account."
                )
            products += [product]
        return products
